/**************************************************************

amass.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "comfun.h"
#include "convert.h"
#include "amass.h"

static int    amass_set = 0;
static double amass_ra  = 0.0;
static double amass_dec = 0.0;
static double amass_lat = 0.0;
static double amass_lon = 0.0;

int airmass_initd(double ra, double de, double lon, double lat)
// Save parameters 
{
	char buf[32], msg[64];
	
  	amass_ra  = ra;
  	amass_dec = de;
  	amass_lon = lon;
  	amass_lat = lat;
  	amass_set = 1;      
  	
  	cmpack_ratostr(amass_ra, buf);
  	sprintf(msg, "R.A. = %s", buf);
  	PrintOut(1, msg);
  	cmpack_ratostr(amass_dec, buf);
  	sprintf(msg, "DEC. = %s", buf);
  	PrintOut(1, msg);
  	cmpack_ratostr(amass_lon, buf);
  	sprintf(msg, "LON. = %s", buf);
  	PrintOut(1, msg);
  	cmpack_ratostr(amass_lat, buf);
  	sprintf(msg, "LAT. = %s", buf);
  	PrintOut(1, msg);
 	
  	return 0;
}

int airmass_inits(const char *ra, const char *dec, const char *lon, const char *lat)
// Convert parameters to double 
{
 double a, d, l, f;
 int stat;

 if (ra) {
	 stat = cmpack_strtora(ra, &a);
	 if (stat!=0) return stat;
 } else {
	 PrintOut(0, "Missing object's right ascension");
	 return MPK_INVALID_PAR;
 }
 if (dec) {
	 stat = cmpack_strtodec(dec, &d);
	 if (stat!=0) return stat;
 } else {
	 PrintOut(0, "Missing object's declination");
	 return MPK_INVALID_PAR;
 }
 if (lon) {
	 stat = cmpack_strtolon(lon, &l);
	 if (stat!=0) return stat;
 } else {
	 PrintOut(0, "Missing observer's longitude");
	 return MPK_INVALID_PAR;
 }
 if (lat) {
	 stat = cmpack_strtolat(lat, &f);
	 if (stat!=0) return stat;
 } else {
	 PrintOut(0, "Missing observer's latitude");
	 return MPK_INVALID_PAR;
 }
 return airmass_initd(a,d,l,f);
}

int airmass(const char *infile, const char *outfile)
{
 FILE  *f, *g;
 char   inbuf[MAXLINE], msg[MAXLINE], *endptr, *aux;
 double jd, x;
 size_t len;

 sprintf(msg, "%s -> %s", infile, outfile);
 PrintOut(0, msg);

 f = fopen(infile, "r");
 if (f==NULL) {
     return MPK_CANNOT_OPEN_SRC;
 }

 g = fopen(outfile, "w");
 if (g==NULL) {
     return MPK_CANNOT_OPEN_OUT;
 }

 while (fgets(inbuf, MAXLINE,f)!=NULL) {
	  // Strip line delimiter(s)
	  aux = strstr(inbuf, "\n"); 
	  if (aux) *aux = '\0'; 
	  aux = strstr(inbuf, "\r"); 
	  if (aux) *aux = '\0'; 
	  // Strip white chars on the end of the line
	  len = strlen(inbuf);
	  if (len>0) {
	  	aux = inbuf + (len-1);
	  	while (*aux==' ' || *aux=='\t') 
	  		*aux-- = '\0';
  	  }
	  // Parse julian date
      jd = strtod(inbuf, &endptr);
      if (strstr(inbuf, "JD ")==inbuf || strstr(inbuf, "JDHEL ")==inbuf || strstr(inbuf, "JDGEO ")==inbuf) {
	      // Header line
    	  fprintf(g, "%s AIRMASS\n", inbuf);
      } else
      if (endptr!=inbuf && jd>=10000 && jd<99999) {
	      // Shortened julian date
	      jd += 2400000.0;
	      x = cmpack_airmass(jd, amass_ra, amass_dec, amass_lon, amass_lat);
	  	  sprintf(msg, "%.5f -> %.3f", jd, x); 
	  	  PrintOut(1, msg);
	  	  fprintf(g, "%s %.3f\n", inbuf, x);
      } else
      if (endptr!=inbuf && jd>=2400000 && jd<3000000) {
	      // Full julian date
	      x = cmpack_airmass(jd, amass_ra, amass_dec, amass_lon, amass_lat);
    	  sprintf(msg, "%.5f -> %.3f", jd, x); 
    	  PrintOut(1, msg);
    	  fprintf(g, "%s %.3f\n", inbuf, x);
      } else {
	      fprintf(g, "%s\n", inbuf);
      }
 }
 fclose(f);
 fclose(g);

 return 0;
}

int airmass_clean(void)
{
	amass_set = 0;
 	return 0;
}
