/**************************************************************

bias.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fitsio.h>

#include "comfun.h"
#include "bias.h"

double bias_minvalue;        // Lowest good datum
double bias_maxvalue;        // Highest good datum
int    bias_nx;              // Width of bias-frame
int    bias_ny;              // Height of bias-frame
double *bias_ddata;          // Data of bias-frame
double *bias_sdata;          // Data of souce-frame
int    bias_overflow;        // Output data were greater than maxvalue
int    bias_underflow;       // Output data were smaller than minvalue

int bias_init(const char *bfile)
// Initializes the variables and loads the bias-frame
{
	FILE *f;
    fitsfile *dfits = NULL;
    long naxes[2];
 	char buf[MAXLINE];
 	int  dstat = 0;
    int anyf, naxis, bitpix, simple, extend;
    long pcount, gcount;

    bias_nx = 0;
    bias_ny = 0;
    bias_ddata = NULL;
    bias_sdata = NULL;
	bias_minvalue = 0.0;
	bias_maxvalue = 65535.0;

    /* Check parameters */
    if (!bfile) {
        PrintOut(0, "Missing name of the bias frame.");
 	    return MPK_INVALID_PAR;
    }
    if (strlen(bfile)==0) {
        PrintOut(0, "Missing name of the bias frame.");
 	    return MPK_INVALID_PAR;
    }
    f = fopen(bfile,"r");
    if (f==NULL) {
 	    PrintOut(1, "Bias frame file not found.");
 	    return MPK_CANNOT_OPEN_BIAS;
    }
    fclose(f);

    /* Open bias file */
    sprintf(buf,"Using bias frame: %s",bfile);
    PrintOut(0, buf);
    
    ffdkopn(&dfits,bfile,READONLY,&dstat);
    if (dstat!=0) return dstat;

    /* Read bias attributes */
	ffghpr(dfits,2,&simple,&bitpix,&naxis,naxes,&pcount,&gcount,&extend,&dstat);
	if (dstat!=0) return dstat;
	bias_nx = naxes[0];
	bias_ny = naxes[1];
	if ((bias_nx<=0)||(bias_ny<=0)) {
		ffclos(dfits,&dstat);
		return MPK_INVALID_SIZE;
	}

	/* Allocate the memory for bias, scientific and output frames */
	bias_ddata = malloc(bias_nx*bias_ny*sizeof(double));
	bias_sdata = malloc(bias_nx*bias_ny*sizeof(double));
	if ((!bias_ddata)||(!bias_sdata)) {
		ffclos(dfits,&dstat);
		return MPK_MEMORY;
	}

	/* Read bias frame data */
    ffg2dd(dfits,1,0.0,bias_nx,bias_nx,bias_ny,bias_ddata,&anyf,&dstat);
	if (dstat!=0) return dstat;

    /* Close bias file */
    ffclos(dfits,&dstat);
    if (dstat!=0) return dstat;

    /* Normal return */
    return 0;
}

int bias(const char *sci, const char *out)
// Computes the output image from source image and bias image (out=sci-bias)
{
 FILE *f;
 int anyf, i, naxis, bitpix, simple, extend, max;
 long naxes[2];
 long pcount, gcount;
 int  sstat = 0, ostat = 0;
 fitsfile *sfits, *ofits;
 char buf[MAXLINE];

 if (out) sprintf(buf,"(%s) => (%s)",sci,out);
     else sprintf(buf,"(%s)",sci);
 PrintOut(0, buf);

  /* Check source file */
  if (!sci) {
      PrintOut(0, "Missing name of the source image file.");
 	  return MPK_INVALID_PAR;
  }
  if (strlen(sci)==0) {
      PrintOut(0, "Missing name of the source image file.");
 	  return MPK_INVALID_PAR;
  }
  f = fopen(sci,"r");
  if (f==NULL) {
      PrintOut(1, "Source image file not found.");
      return MPK_CANNOT_OPEN_SRC;
  }
  fclose(f);

 /* Open exposure file */
 ffdkopn(&sfits,sci,(out?READONLY:READWRITE),&sstat);
 if (sstat!=0) return sstat;

 /* Read image attributes */
 ffghpr(sfits,2,&simple,&bitpix,&naxis,naxes,&pcount,&gcount,&extend,&sstat);
 if (sstat!=0) return sstat;
 if ((naxes[0]<=0)||(naxes[1]<=0)) {
	 ffclos(sfits,&sstat);
     return MPK_INVALID_SIZE;
 }
 if ((bias_nx!=naxes[0])||(bias_ny!=naxes[1])) {
	 ffclos(sfits,&sstat);
	 return MPK_DIFF_BIAS_SIZE;
 }

  /* Get equivalent data type and determine the max. and min. value */ 
  ffgiet(sfits,&bitpix,&sstat);
  switch (bitpix) 
  {
  case BYTE_IMG:  
    bias_minvalue = 0.0;
    bias_maxvalue = 255.0;
    break;
  case SBYTE_IMG:  
    bias_minvalue = -128.0;
    bias_maxvalue = 127.0;
    break;
  case SHORT_IMG:
    bias_minvalue = -32768.0;
    bias_maxvalue = 32767.0;
    break;
  case USHORT_IMG:
    bias_minvalue = 0.0;
    bias_maxvalue = 65535.0;
    break;
  case LONG_IMG:
    bias_minvalue = -2147483648.0;
    bias_maxvalue = 2147483647.0;
    break;
  case ULONG_IMG:
    bias_minvalue = 0.0;
    bias_maxvalue = 4294967296.0;
    break;
  default:
    bias_minvalue = 0.0;
    bias_maxvalue = 65535.0;
  }
	sprintf(buf, "Min. value   : %.1f", bias_minvalue); 
	PrintOut(1, buf);
	sprintf(buf, "Max. value   : %.1f", bias_maxvalue); 
	PrintOut(1, buf);
 
 /* Read exposure data */
 ffg2dd(sfits,1,0.0,bias_nx,bias_nx,bias_ny,bias_sdata,&anyf,&sstat);
 if (sstat!=0) {
	 ffclos(sfits,&sstat);
	 return sstat;
 }

 /* Counts a range of intensities of the image */
 sstat = 0;

 /* bias subtraction and range checking */
 max = naxes[0]*naxes[1];
 bias_underflow = bias_overflow = 0;
 for (i=0;i<max;i++) {
	  bias_sdata[i]=bias_sdata[i]-bias_ddata[i];
	  if (bias_sdata[i]<bias_minvalue) {
		  bias_sdata[i] = bias_minvalue;
		  bias_underflow = 1;
	  }
	  if (bias_sdata[i]>bias_maxvalue) {
		  bias_sdata[i] = bias_maxvalue;
		  bias_overflow = 1;
	  }
 }

 if (out) {
 	/* output initialisation - parameters by scientific image */
 	remove(out);
 	ffdkinit(&ofits,out,&ostat);
 	if (ostat!=0) return MPK_CANNOT_OPEN_OUT;
 	ffcopy(sfits,ofits,1,&ostat);
 	ffphis(ofits,"Biasframe subtracted.",&ostat);
 	/* corrected image output */
 	ffp2dd(ofits,1,bias_nx,bias_nx,bias_ny,bias_sdata,&ostat);
 	if (ostat!=0) return ostat;
	/* Close output file */
 	ffclos(ofits,&ostat);
 	if (ostat!=0) return ostat;
 } else {
	 /* overwrite input image */
   	 ffphis(sfits,"Biasframe subtracted.",&sstat);
  	 ffp2dd(sfits,1,bias_nx,bias_nx,bias_ny,bias_sdata,&sstat);
  	 if (sstat!=0) return sstat;
 }

  /* Close exposure file */
  ffclos(sfits,&sstat);
  if (sstat!=0) return sstat;

  /* Normal return */
  if (bias_overflow) {
	  sprintf(buf,"An overflow has been occurred during computation (max.=%.1f).", bias_maxvalue);
	  PrintOut(1, buf);
  }
  if (bias_underflow) {
   	  sprintf(buf,"An underflow has been occurred during computation (min.=%.1f).", bias_minvalue);
	  PrintOut(1, buf);
  }
  return 0;
}

int bias_clean(void)
// Frees allocated memory
{
 if (bias_ddata) free(bias_ddata);
 if (bias_sdata) free(bias_sdata);
 return 0;
}
