/**************************************************************

autoflat.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fitsio.h>

#ifdef _WIN32
#include <io.h>
#endif

#include "cmunipack.h"
#include "comfun.h"
#include "kombine.h"
#include "robmean.h"

void PrintHelp()
{
 printf("KOMBINE version %s - utility for the merging CCD frames.\n",VERSION);
 printf(PROJECTLINE);
 printf("\nUsage:kombine [options] [@dirfile] [out=file] [image1] ...\n\n");
 printf("parameters:\n");
 printf("     @dirfile=  file with input image names\n");
 printf("                (for only '@', read image names from standart input)\n");
 printf("     out=       output file name (default = %s)\n",MASKNAME);
 printf("     image1...  name(s) of simple input image(s)\n");
 printf("options:\n");
 printf("     -o<file>   output file name (equivalent with 'out=')\n");
 printf("     -q         quiet mode\n");
 printf("     -h         give this help\n");
 printf("     -v         print software version\n");
 printf("     -l         display software license\n");
 printf(WARRANTYLINE);
 printf(HOMEPAGELINE);
 printf(BUGREPORTS);
}

void report(int status, const char *file)
{
 if (cmpack_debug) 
 	fprintf(stderr, "The following error(s) has been occured: %d\n",status);
 if (status<1000) 
 	ffrprt(stderr, status);
 else  
 	PrintError(PROGNAME, file, 0, status);
}

void error(int status, const char *file)
{
 report(status, file);
 exit(status);
}

int main(int argc, char **argv)
{
    int stdinput = FALSE, i, res, level = 0;
    char line[MAXLINE], buf1[MAXLINE];
    char *par, *out = MASKNAME, *namsci = "";
	FILE *f;

	/* Zpracovani parametru prikazove radky */
	for (i=1;i<argc;i++) {
	   par = argv[i];

	   /* Prints help */
	   if ((strcmp(par,"-h")==0)||(strcmp(par,"--help")==0)) {
		  PrintHelp();
          return 0;
       }
	   /* Prints version */
	   if ((strcmp(par,"-v")==0)||(strcmp(par,"--version")==0)) {
		  PrintVersion(PROGNAME);
          return 0;
       }
       /* Prints license */
       if ((strcmp(par,"-l")==0)||(strcmp(par,"--license")==0)) {
          PrintLicense();
          return 0;
       }
       /* Flag for input from stdin or dirfile */
       if (strchr(par,'@')==par) {
	       namsci = malloc(strlen(par));
	       strcpy(namsci,&par[1]);
           stdinput = strlen(namsci)==0;
       }
       /* Name of output file */
       if (strstr(par,"out=")==par) {
	       out = malloc(strlen(par)-3);
	       strcpy(out,&par[4]);
       }
       if (strstr(par,"-o")==par) {
	       out = malloc(strlen(par)-1);
	       strcpy(out,&par[2]);
       }
       /* Quiet/verbose flags */
       if (strcmp(par,"-q")==0 || strcmp(par,"--quiet")==0 || strcmp(par,"--silent")==0) {
           level = MPK_LEVEL_QUIET;
       }
       if (strcmp(par,"--verbose")==0 || strcmp(par,"--debug")==0) {
           level = MPK_LEVEL_DEBUG;
       }
  }
  
  cmpack_setoutput(PrintMessage, level);
  PrintPrologue(PROGNAME);
  res=kombine_init(NULL);
  if (res!=0) error(res, NULL);
  res=kombine_open(out);
  if (res!=0) error(res, NULL);

  /* image names reads from command line */
  for (i=1;i<argc;i++) {
	  par = argv[i];
      if ((strchr(par,'=')==0)&&(strchr(par,'@')!=par)&&(strchr(par,'-')!=par)) {
#ifdef _WIN32
		intptr_t fh;
    	struct _finddata_t fd;
		if ((fh=_findfirst(par, &fd))!=-1) {
			do {
				if ((fd.attrib & _A_SUBDIR)==0)  {
					res = kombine(fd.name);
          			if (res!=0) report(res, fd.name);
          		}
          	} while (_findnext(fh, &fd)==0);
			_findclose(fh);
      	 } else {
	      	 report(MPK_CANNOT_OPEN_SRC, par);
      	 }
#else
        res = kombine(par);
		if (res!=0) report(res, par);
#endif
      }
  }

  /* image names reads from dirfile */
    if (strlen(namsci)>0) {
	f = fopen(namsci,"r");
	if (f==NULL) 
	  	error(MPK_NO_DIRFILE, namsci);
    while (fgets(buf1,MAXLINE,f)!=NULL) {
        sscanf(buf1,"%s",line);
        if (strcmp(line,"")) {
	        res = kombine(line);
            if (res!=0) report(res, line);
        }
    }
  }

  /* Image names reads from stdin */
  if (stdinput) {
      while (fgets(buf1,MAXLINE,stdin)!=NULL) {
          sscanf(buf1,"%s",line);
          if (strcmp(line,"")) {
            res = kombine(line);
   	        if (res!=0) report(res, line);
          }
      }
  }

  /* Write output */
  res=kombine_close(); 
  if (res!=0) report(res, NULL);

  res=kombine_clean();
  if (res!=0) error(res, NULL);

  return 0;
}
