/**************************************************************

iost.c (C-Munipack project)
SBIG files low-level routines
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "comfun.h"
#include "iost.h"

struct __stfile
{
	FILE *f;
	char head[STHEADSIZE];
	unsigned short int *data;
};

typedef struct __stfile __stfile;

int stgetftype(const char *head)
/* Compressed or uncompressed file? */
{
  char *aux, *aux2;

  if ((strstr(head,"ST-")==head)||(strstr(head,"SBIG")==head)||(strstr(head,"PixCel")==head)) {
      aux  = strchr(head,0x0A);
	  aux2 = strstr(head,"Compressed");
	  if (aux!=NULL) {
		  /* SBIG compressed or uncompressed file */
          return ((aux2!=NULL)&&(aux2<aux)?2:1);
	  }
  }
  /* Not SBIG file */
  return -1;
}

void stopen(stfile **fs, const char *filename, int *stat)
/* Open the file and read the header into memory */
{
  __stfile *st;
  size_t headlen;

  st = (__stfile *)malloc(sizeof(__stfile));
  if (st==NULL) {
	  *stat = MPK_MEMORY;                   /* Cannot allocate struct stfile */
	  *fs = NULL;
  	  return;
  }
  memset(st,0,sizeof(__stfile));

  /* open the file */
  st->f = fopen(filename,"rb");
  if (st->f==NULL) {
	  stclos(st, stat);
	  *fs = NULL;
	  *stat = MPK_CANNOT_OPEN_SRC;          /* Cannot open file */
	  return;
  }

  /* Read header of the file */
  headlen = fread(st->head,1,STHEADSIZE,st->f);
  st->data = NULL;

  /* Identify the file type */
  if ((headlen!=STHEADSIZE)||(stgetftype(st->head)<0)) {
	  stclos(st, stat);
	  *fs = NULL;
	  *stat = MPK_INVALID_HEADER;          /* Cannot identify the type of the file */
	  return;
  }

  /* Normal return */
  *fs = st;
  *stat = 0;
}

void stclos(stfile *stfile, int *stat)
/* Close a file and free allocated memory */
{
 __stfile *st = (__stfile *)stfile;
 if (st) {
    if (st->f)    fclose(st->f);
    if (st->data) free(st->data);
    free(st);
 }
 *stat = 0;
}

void stgkys(stfile *stfile, const char *key, char *buf, int maxlen, int *stat)
/* Get string by key */
{
 char *aux, *j;
 __stfile *st = (__stfile *)stfile;

 buf[0] = 0;
 if (st==NULL) {
	 *stat = MPK_CLOSED_FILE;        /* The file is not open */
	 return;
 }
 /* Locate a key in the header */
 aux = (char *)malloc(strlen(key)+3);
 if (aux==NULL) {
	 *stat = MPK_MEMORY;
	 return;
 }
 strcpy(aux,key);
 strcat(aux," =");
 j = strstr(st->head,aux);
 free(aux);
 if (j==NULL) {
	 *stat = MPK_KEY_NOT_FOUND;      /* The key was not found */
	 return;
 }
 /* Copy the value to buf */
 j = j + strlen(key) + 2;
 if (buf) sscanf(j,"%s",buf);
 *stat = 0;
}

void stgkyj(stfile *stfile, const char *key, long *value, int *stat)
/* Get integer number by key */
{
 char *aux, *j;
 __stfile *st = (__stfile *)stfile;

 *value = 0;
 if (st==NULL) {
	 *stat = MPK_CLOSED_FILE;        /* The file is not open */
	 return;
 }
 /* Locate a key in the header */
 aux = (char *)malloc(strlen(key)+3);
 if (aux==NULL) {
	 *stat = MPK_MEMORY;
	 return;
 }
 strcpy(aux,key);
 strcat(aux," =");
 j = strstr(st->head,aux);
 free(aux);
 if (j==NULL) {
	 *stat = MPK_KEY_NOT_FOUND;      /* The key was not found */
	 return;
 }
 /* Convert a value */
 j = j + strlen(key) + 2;
 *value = atol(j);
 *stat = 0;
}

void stgkyd(stfile *stfile, const char *key, double *value, int *stat)
/* Get floating number by key. */
{
 char *aux, *j;
 __stfile *st = (__stfile *)stfile;

 *value = 0.0;
 if (st==NULL) {
	 *stat = MPK_CLOSED_FILE;        /* The file is not open */
	 return;
 }
 /* Locate a key in the header */
 aux = (char *)malloc(strlen(key)+3);
 if (aux==NULL) {
	 *stat = MPK_MEMORY;
	 return;
 }
 strcpy(aux,key);
 strcat(aux," =");
 j = strstr(st->head,aux);
 free(aux);
 if (j==NULL) {
	 *stat = MPK_KEY_NOT_FOUND;      /* The key was not found */
	 return;
 }
 /* Convert a value */
 j = j + strlen(key) + 2;
 *value = atof(j);
 *stat = 0;
}

int stgkyn(stfile *stfile,  int nkey, char *keyname, char *keyval, char *comm, int *stat) {
/* Get key name and value by index. Comm is set always to "" */
 char *aux, *j, buf[MAXLINE];
 __stfile *st = (__stfile *)stfile;

 if (keyname) keyname[0] = '\0';
 if (keyval)  keyval[0]  = '\0';
 if (comm)    comm[0]    = '\0';
 if (st==NULL) {
	 *stat = MPK_CLOSED_FILE;        /* The file is not open */
	 return -1;
 }
 /* Locate a nth line in the header */
 aux = st->head;
 while ((aux!=NULL)&&(nkey>0)) {
 	j = strchr(aux,0x0A);
	if (j) {
		aux = j+2;
		nkey--;
	}
 }
 if (aux==NULL) {
	 *stat = MPK_KEY_NOT_FOUND;
	 return -1;
 }
 /* Extract the line from the header */
 j = strchr(aux,0x0A);
 strncpy(buf,aux,j-aux);
 buf[j-aux] = '\0';
 if (strcmp(buf,"End")==0) {
	 *stat = MPK_KEY_NOT_FOUND;
	 return -1;
 }
 /* Parse buf into keyname, keyval (comm is always empty) */
 j = strchr(buf,'=');
 if ((j!=NULL)&&(keyval)) {
    strncpy(keyname,buf,j-buf-1);
    keyname[j-buf-1] = '\0';
    strncpy(keyval,j+2,64);
 } else
 if (keyname)
    strncpy(keyname,buf,64);
 return 0;
}

void stghpr(stfile *stfile, int *nx, int *ny, int *stat)
{
 __stfile *st = (__stfile *)stfile;
 long x, y;

 *nx = *ny = 0;
 if (st==NULL) {
	 *stat = MPK_CLOSED_FILE;        /* The file is not open */
	 return;
 }
 /* Get width and height of the image */
 stgkyj(st,"Width",&x,stat);
 stgkyj(st,"Height",&y,stat);
 if ((*stat!=0)||(x<=0)||(y<=0)||(x>=100000)||(y>=100000)) {
	 *stat = MPK_INVALID_HEADER;        /* Invalid image size */
 } else {
 	 *nx = x; *ny = y;
 	 *stat = 0;
 }
}

void stg2dd(stfile *stfile, int flipx, int flipy, double *ccd, int *stat)
{
 __stfile *st = (__stfile *)stfile;
 long nx, ny;
 int i, j, ii, jj, k, deltai, hodnota;
 size_t headlen;
 unsigned short delka;
 unsigned char *buf;

 if (st==NULL) {
	 *stat = MPK_CLOSED_FILE;        /* The file is not open */
	 return;
 }
 if (st->data==NULL) {
 	/* Get width and height of the image */
 	stgkyj(st,"Width",&nx,stat);
 	stgkyj(st,"Height",&ny,stat);
 	if ((nx<=0)||(ny<=0)||(nx>=100000)||(ny>=100000)) {
     	*stat = MPK_INVALID_HEADER;     /* Cannot read the width or height of the image */
     	return;
 	}
 	/* Allocate internal buffer */
 	st->data = (unsigned short *)malloc(nx*ny*sizeof(unsigned short));
 	if (st->data==NULL) { *stat = MPK_MEMORY; return; }
 	/* Read image data */
 	fseek(st->f, STHEADSIZE, SEEK_SET);
 	if (stgetftype(st->head)!=2) {
    	 /* uncompressed image */
		 size_t bytes = nx*ny;
	 	 if (fread(st->data,2,bytes,st->f)!=bytes) {
     	 	*stat = MPK_INVALID_DATA;                            /* invalid image format */
         	return;
	 	}
  	} else {
	  	/* Compressed image, courtesy by P. Pravec */
      	buf = (unsigned char *)malloc(2*nx);
      	if (buf==NULL) { *stat = MPK_MEMORY; return; }
      	for (i=0;i<ny;i++) {   /* for each line */
          	/* length of line in bytes */
          	headlen = fread(&delka,1,2,st->f);
          	if ((headlen<2)||(delka>2*nx)) {
	          	/* invalid row length */
	          	*stat = MPK_INVALID_DATA;                      /* invalid image format */
              	free(buf);
              	return;
          	} else
          	if (delka==2*nx) {
	          	/* no compressed line in compressed image */
	          	headlen = fread(&st->data[i*nx],1,2*nx,st->f);
	          	if ((int)headlen!=2*nx) {
   	              	/* invalid row length */
	              	*stat = MPK_INVALID_DATA;                            /* invalid image format */
	              	free(buf);
                  	return;
                }
          	} else {
	          	/* compressed line */
              	headlen = fread(buf,1,delka,st->f);
	          	if (headlen!=delka) {
   	              	/* invalid row length */
	              	*stat = MPK_INVALID_DATA;                            /* invalid image format */
	              	free(buf);
                  	return;
              	}
              	hodnota = buf[0] + 256*buf[1]; /* first value in the line */
              	st->data[i*nx] = hodnota;
              	j = 1; k = 2;
              	while (k<delka) {
	              	deltai = buf[k];
	              	k++;
	              	if (deltai>127) deltai -= 256;
                  	if (deltai==-128) {
	                  	hodnota=buf[k]+256*buf[k+1];
	                  	k+=2;
                  	} else {
	                  	hodnota +=deltai;
                  	}
                  	st->data[i*nx+j] = hodnota;
                  	j++;
              	}
          	}
       }
   }
 }
 /* Copy values from data to buf (image flipping) */
 for (j=0;j<ny;j++) {
     for (i=0;i<nx;i++) {
	     if (flipx) ii = nx-i-1; else ii = i;
		 if (flipy) jj = ny-j-1; else jj = j;
		 ccd[j*nx+i] = st->data[jj*nx+ii];
	 }
 }
 *stat = 0;
}

/* Reads observation date */
void stgetdate(void *st, int *yr, int *mon, int *day, int *stat)
{
    long y=0, y2k=0, m=0, d=0;
    char datestr[MAXLINE];

	stgkys(st, "Date", datestr, MAXLINE, stat);
	stgkyj(st, "Y2KYear", &y2k, stat);
    if (*stat!=0) y2k = 0;
    /* Parse input date and time */
    sscanf(datestr, "%2ld%*1s%2ld%*1s%ld", &m, &d, &y);
    /* Compute the year */
	*mon = m;
	*day = d;
    if (y2k>0)
        *yr = y2k;
    else if ((y>50)&&(y<=999))
        *yr = y + 1900;
    else if ((y>=0)&&(y<50))
        *yr = y + 2000;
    else if ((y>=1900)&&(y<1950))
        *yr = y + 100;
    else
        *yr = y;
	/* Check values */
 	if ((*yr>1900)&&(*yr<=9999)&&(*mon>=1)&&(*mon<=12)&&(*day>=1)&&(*day<=31))
    	*stat = 0;
	else
		*stat = MPK_INVALID_DATE;
}

/* Reads observation time */
void stgettime(void *st, int *hr, int *min, int *sec, int *stat)
{
    char timestr[MAXLINE];

	stgkys(st, "Time", timestr, MAXLINE, stat);
    if (*stat!=0) strcpy(timestr,"");
    /* Parse input date and time */
    sscanf(timestr,"%2d%*1s%2d%*1s%2d",hr,min,sec);
	/* Check values */
	if ((*hr>=0)&&(*hr<=23)&&(*min>=0)&&(*min<=59)&&(*sec>=0)&&(*sec<=59))
    	*stat = 0;
	else
		*stat = MPK_INVALID_DATE;
}
