/**************************************************************

konvfits.c (C-Munipack project)
Conversion functions for FITS files
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fitsio.h>

#include "comfun.h"
#include "convert.h"
#include "fitsio.h"
#include "trajd.h"
#include "konvfits.h"
#include "fitsfun.h"

typedef struct {
	fitsfile *fits;
	int status;
} fitstruct;

/* Open FITS file */
int fits_open(void **fits, const char *filename)
{
	fitsfile *f;
	int st = 0;
	fitstruct *fs;

	/* Open file */
	ffdkopn(&f, filename, READONLY, &st);
	if (st!=0) {
		*fits = NULL;
		return st;
	}
	/* Create descriptor */
	fs = (fitstruct*) malloc(sizeof(fitstruct));
	if (fs==NULL) {
		*fits = NULL;
		return MPK_MEMORY;
	}
	fs->fits = f;
	fs->status = st;
	*fits = fs;
	return 0;
}

/* Close file and free allocated memory */
int fits_close(void *fits)
{
	fitstruct *fs;

	if (fits==NULL)
		return MPK_CLOSED_FILE;

	fs = (fitstruct*) fits;
	ffclos(fs->fits, &fs->status);
	free(fs);
	return 0;
}

/* Reads string value from header */
int fits_gets(void *fits, const char *key, char *buf, int buflen)
{
	char val[FLEN_VALUE];
	int a, b, c;
	fitstruct *fs;
	
	buf[0] = 0;
	if (fits==NULL) return MPK_CLOSED_FILE;
	fs = (fitstruct*) fits;
	fs->status = 0;
	/* Virtual keys */
	if (strcmp(key, "_magic_")==0) {
		strncpy(buf, "FITS (Flexible Image Transport System)", buflen);
		return 0;
	} else
	if (strcmp(key, "_filter_")==0) {
		fs->status = fits_gfilter(fs->fits, val);
		if (fs->status==0) strncpy(buf, val, buflen);
		return fs->status;
	} else
	if (strcmp(key, "_object_")==0) {
		ffgkys(fs->fits, "OBJECT", val, NULL, &fs->status);
		if (fs->status==0) strncpy(buf, val, buflen);
		return fs->status;
	} else
	if (strcmp(key, "_date_")==0) {
	    fs->status = fits_getdatetime(fs->fits, val, NULL);
		if (fs->status==0) strncpy(buf, val, buflen);
		return fs->status;
	} else
	if (strcmp(key, "_time_")==0) {
	    fs->status = fits_getdatetime(fs->fits, NULL, val);
		if (fs->status==0) strncpy(buf, val, buflen);
		return fs->status;
	} else
	if (strcmp(key, "_ra_")==0) {
	    fs->status = fits_getra(fs->fits, &a, &b, &c);
		if (fs->status==0) {
			cmpack_ratostr(a+b/60.0+c/3600.0, val);
			strncpy(buf, val, buflen);
		}
		return fs->status;
	} else
	if (strcmp(key, "_dec_")==0) {
	    fs->status = fits_getdec(fs->fits, &a, &b, &c);
		if (fs->status==0) {
			cmpack_dectostr(a+b/60.0+c/3600.0, val);
			strncpy(buf, val, buflen);
		}
		return fs->status;
	} else {
		/* Physical keys */
		ffgkys(fs->fits, (char*) key, val, NULL, &fs->status);
		if (fs->status==0) strncpy(buf, val, buflen);
		return fs->status;
	}
}

/* Reads integer value from header */
int fits_getl(void *fits, const char *key, long *val)
{
	int simple, extend, naxis, bitpix, fravg;
    long pcount, gcount, naxes[2];
	fitstruct *fs;

    *val = 0;
	if (fits==NULL) return MPK_CLOSED_FILE;
	fs = (fitstruct*) fits;
	fs->status = 0;
	/* Virtual keys */
	if (strcmp(key, "_width_")==0) {
    	ffghpr(fs->fits, 2, &simple, &bitpix, &naxis, naxes, &pcount, &gcount, &extend, &fs->status);
		if (fs->status==0) *val = naxes[0];
		return fs->status;
	} else
	if (strcmp(key, "_height_")==0) {
    	ffghpr(fs->fits, 2, &simple, &bitpix, &naxis, naxes, &pcount, &gcount, &extend, &fs->status);
		if (fs->status==0) *val = naxes[1];
		return fs->status;
	} else
	if (strcmp(key, "_bitcount_")==0) {
    	ffghpr(fs->fits, 2, &simple, &bitpix, &naxis, naxes, &pcount, &gcount, &extend, &fs->status);
		if (fs->status==0) *val = bitpix;
		return fs->status;
	} else 
	if (strcmp(key, "_fravg_")==0) {
    	fs->status = fits_gfravg(fs->fits, &fravg);
		if (fs->status==0) *val = fravg;
		return fs->status;
	} else {
		/* Physical keys */
		ffgkyj(fs->fits, (char*) key, val, NULL, &fs->status);
		return fs->status;
	}
}

/* Reads floating point value from header */
int fits_getd(void *fits, const char *key, double *val)
{
	int a, b, c;
	fitstruct *fs;
	
    *val = 0.0;	
	if (fits==NULL) return MPK_CLOSED_FILE;
	fs = (fitstruct*) fits;
	/* Virtual keys */
	if (strcmp(key, "_temperature_")==0) {
		return fits_gtemp(fs->fits, val);
	} else
	if (strcmp(key, "_exposure_")==0) {
		return fits_gexp(fs->fits, val);
	} else
	if (strcmp(key, "_background_")==0) {
		return fits_gblack(fs->fits, val);
	} else
	if (strcmp(key, "_range_")==0) {
		return fits_grange(fs->fits, val);
	} else
	if (strcmp(key, "_ra_")==0) {
	    fs->status = fits_getra(fs->fits, &a, &b, &c);
		if (fs->status==0) *val = a+b/60.0+c/3600.0;
		return fs->status;
	} else
	if (strcmp(key, "_dec_")==0) {
	    fs->status = fits_getdec(fs->fits, &a, &b, &c);
		if (fs->status==0) *val = a+b/60.0+c/3600.0;
		return fs->status;
	} else {
		/* Physical keys */
		ffgkyd(fs->fits, (char*) key, val, NULL, &fs->status);
		return fs->status;
	}
}

int fits_gkyn(void *fits, int index, char *key, int keylen, char *val, int vallen, char *com, int comlen)
/* Random access reading of (key, value, comment) triples from hedader */
{
	char k[FLEN_KEYWORD], v[FLEN_VALUE], c[FLEN_COMMENT];
	fitstruct *fs;

	if (fits==NULL)
		return MPK_CLOSED_FILE;

	fs = (fitstruct*) fits;
	fs->status = 0;
	ffgkyn(fs->fits, index, k, v, c, &fs->status);
	if (fs->status==0) {
		strncpy(key, k, keylen);
		strncpy(val, v, vallen);
		strncpy(com, c, comlen);
	}
	return fs->status;
}

/* Reads image data to buffer of floating points */
int fits_g2dd(void *fits, double *buf, int flipx, int flipy)
{
	int simple, extend, naxis, bitpix, anyf;
    long pcount, gcount, naxes[2];
	fitstruct *fs;

	if (fits==NULL)
		return MPK_CLOSED_FILE;
	if (buf==NULL)
		return MPK_MEMORY;

	fs = (fitstruct*) fits;
	fs->status = 0;
    ffghpr(fs->fits, 2, &simple, &bitpix, &naxis, naxes, &pcount, &gcount, &extend, &fs->status);
	ffg2dd(fs->fits, 1, 0, naxes[0], naxes[0], naxes[1], buf, &anyf, &fs->status);
	if (flipx || flipy)
		fits_flipimage(fs->fits, flipx, flipy, &fs->status);
	return fs->status;
}

