/**************************************************************

konvoes.c (C-Munipack project)
Conversion functions for OES Astro files
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fitsio.h>

#include "comfun.h"
#include "fitsfun.h"
#include "trajd.h"
#include "konve.h"
#include "konvoes.h"
#include "oesfile.h"

#define SATURATE   4096.0
#define BITPIX     -32
#define BIAS       0.0
#define BSCALE     1.0

/* Returns nonzero if the file is in oes format */
int oes_test(FILE *f)
{
  size_t size;
  unsigned long buf[64];

  rewind(f);
  size = fread(&buf[0], 4, 64, f);
  if ((size==64)&&(((buf[62]==H_MAG1_OWN)&&(buf[63]==H_MAG2_OWN))||((buf[62]==H_MAG1_INV)&&(buf[63]==H_MAG2_INV))))
  		return 1;
  else
  		return 0;
}

/* Read parameters from SBIG file */
int oes_open(void **handle, const char *filename)
{
	return oesopen(handle, filename);
}

/* Close file and free allocated memory */
int oes_close(void *handle)
{
	oesclos(handle);
	return 0;
}

/* Reads string value from header */
int oes_gets(void *handle, const char *key, char *buf, int buflen)
{
	int yr, mo, dy, hr, mi, se;

	buf[0] = 0;
	if (handle==NULL)
		return MPK_CLOSED_FILE;
	/* Virtual keys */
	if (strcmp(key, "_magic_")==0) {
		strncpy(buf, "OES Astro file", buflen);
		return 0;
	} else
	if (strcmp(key, "_date_")==0) {
 		if (oesgdat(handle, &yr, &mo, &dy, &hr, &mi, &se)==0) {
			sprintf(buf, "%04d-%02d-%02d", yr, mo, dy);
			return 0;
		} else
			return MPK_INVALID_DATE;
	} else
	if (strcmp(key, "_time_")==0) {
 		if (oesgdat(handle, &yr, &mo, &dy, &hr, &mi, &se)==0) {
			sprintf(buf, "%02d:%02d:%02d", hr, mi, se);
			return 0;
		} else
			return MPK_INVALID_DATE;
	} else {
	    /* Physical keys */
	    return oesgkys(handle, atoi(key), buf, buflen);
    }
}

/* Reads integer value from header */
int oes_getl(void *handle, const char *key, long *value)
{
	int nx, ny;

	*value = 0;
	if (handle==NULL)
		return MPK_CLOSED_FILE;
	/* Virtual keys */
	if (strcmp(key, "_width_")==0) {
		if (oesghpr(handle, &nx, &ny)==0) {
			*value = nx;
			return 0;
		} else
			return MPK_INVALID_HEADER;
	} else
	if (strcmp(key, "_height_")==0) {
		if (oesghpr(handle, &nx, &ny)==0) {
			*value = ny;
			return 0;
		} else
			return MPK_INVALID_HEADER;
		return 0;
	} else
	if (strcmp(key, "_bitpix_")==0) {
		*value = 12;
		return 0;
	} else {
		/* Physical keys */
		return oesgkyj(handle, atoi(key), value);
	}
}

/* Reads floating point value from header */
int oes_getd(void *handle, const char *key, double *value)
{
	*value = 0.0;
	if (handle==NULL)
		return MPK_CLOSED_FILE;
	/* Virtual keys */
	if (strcmp(key, "_exposure_")==0) {
		return oesgexp(handle, value);
	} else
	if (strcmp(key, "_temperature_")==0) {
		return oesgtem(handle, value);
	} else {
		/* Physical keys */
		return oesgkyd(handle, atoi(key), value);
	}
}

int oes_convhdr(void *oes, fitsfile *fits)
/* Converts header from oes file to FITS file */
{
  long naxes[2];
  int sstat, dstat=0, nx, ny, yr, mo, dy, hr, mi, se;
  double exp, temp;
  char actdate[MAXLINE], buf[MAXLINE];
  time_t tsec;
  struct tm *actim;

  /* Set image dimensions */
  sstat = oesghpr(oes, &nx, &ny);
  if (sstat!=0) return sstat;
  naxes[0]=nx; naxes[1]=ny;
  ffphps(fits,BITPIX,2,naxes,&dstat);

  /* Set date and time of observation */
  sstat = oesgdat(oes, &yr, &mo, &dy, &hr, &mi, &se);
  if (sstat!=0) return sstat;
  fits_setdt(fits, yr, mo, dy, hr, mi, se, 0);

  /* Set actual date and time */
  tsec = time(&tsec);
  actim = localtime(&tsec);
  sprintf(actdate,"%04d-%02d-%02d",actim->tm_year+1900,actim->tm_mon+1,actim->tm_mday);
  ffpkys(fits,"DATE",actdate,"DATE OF CREATION",&dstat);

  sprintf(buf,"%s (%s) %s", PROGNAME, PKGNAME, VERSION);
  ffpkys(fits,"ORIGIN",buf,"",&dstat);
  ffpkyf(fits,"BSCALE",BSCALE,1," ",&dstat);
  ffpkyf(fits,"BZERO",BIAS,1," ",&dstat);

  sstat = oesgexp(oes, &exp);
  if (sstat==0) ffpkyg(fits,"EXPTIME",exp,2,"EXPOSURE IN SECONDS",&dstat);
  sstat = oesgtem(oes, &temp);
  if (sstat==0) ffpkyg(fits,"TEMPERAT",temp,1,"TEMPERATURE IN DEGREES C",&dstat);

//  ffpkys(fits,"FILTER","","OPTICAL FILTER NAME",&dstat);
  return 0;
}

int oes_convimg(void *oes, fitsfile *fits, int flipx, int flipy)
/* Converts image data from oes file to FITS file */
{
  double *image;
  int sstat, dstat=0, nx, ny;

  /* Get image size */
  sstat = oesghpr(oes, &nx, &ny);
  if (sstat!=0) return sstat;

  /* Copy image data */
  image = (double*)malloc(nx*ny*sizeof(double));
  sstat = oesg2dd(oes,flipx,flipy,SATURATE,image);
  if (sstat!=0) { free(image); return sstat; }
  ffp2dd(fits,1,nx,nx,ny,image,&dstat);
  if (dstat!=0) { free(image); return dstat; }

  /* Normal return */
  free(image);
  return 0;
}

int oes_gkyn(void *handle, int nkey, char *key, int keylen, char *val, int vallen, char *com, int comlen)
     // Sequential read of parameters
{
	return oesgkyn(handle, nkey, key, val, com);
}

int oes_g2dd(void *handle, double *buf, int flipx, int flipy)
{
	return oesg2dd(handle, flipx, flipy, SATURATE, buf);
}
