/**************************************************************

konvsbig.c (C-Munipack project)
Conversion functions for SBIG files
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fitsio.h>

#include "comfun.h"
#include "fitsfun.h"
#include "convert.h"
#include "iost.h"
#include "trajd.h"
#include "konve.h"
#include "konvsbig.h"

#define BIAS       	32768
#define HEADSIZE  	2048             /* header length of the SBIG image file */

typedef struct
{
	stfile *st;
	int status;
} st_file;

/* Returns 1 if the file is in SBIG format */
int st_test(FILE *f)
{
  int size;
  char buf[64];
  // Get size of file
  fseek(f,0,SEEK_END);
  size = ftell(f);
  if (size>HEADSIZE) {
	  // Read first 32 bytes
	  rewind(f);
      if (fread(buf,32,1,f)) {
          if (((strstr(buf,"ST-")==buf)||(strstr(buf,"SBIG")==buf)||(strstr(buf,"PixCel")==buf))
		  		&&((strchr(buf,0x0A)!=NULL))) {
  		      return 1;
	      }
      }
  }
  return 0;
}

int st_open(void **handle, const char *filename)
/* Open the file and read the header into memory */
{
	int status = 0;
	st_file *fd;
	stfile *st;

	stopen(&st, filename, &status);
	if (status!=0) return status;
	fd = (st_file*) malloc(sizeof(st_file));
	fd->st = st;
	fd->status = status;
	*handle = fd;
	return 0;
}

int st_close(void *handle)
/* Close a file and free allocated memory */
{
	st_file *st;

	if (handle==NULL) return MPK_CLOSED_FILE;
	st = (st_file*) handle;
	stclos(st->st, &st->status);
    free(st);
	return 0;
}

/* Reads string value from header */
int st_gets(void *handle, const char *key, char *buf, int buflen)
{
	int yr, mo, dy, hr, mi, se;
	st_file *st;

 	buf[0] = 0;
	if (handle==NULL) return MPK_CLOSED_FILE;	/* Invalid handle */
	st = (st_file *) handle;
	/* Virtual keys */
	if (strcmp(key, "_filter_")==0) {
		stgkys(st->st, "Filter", buf, buflen, &st->status);
		return st->status;
	} else
	if (strcmp(key, "_date_")==0) {
		if (buflen<11) return MPK_INVALID_PAR;
		stgetdate(st->st, &yr, &mo, &dy, &st->status);
		if (st->status!=0) return st->status;
		sprintf(buf, "%04d-%02d-%02d", yr, mo, dy);
		return 0;
	} else
	if (strcmp(key, "_time_")==0) {
		if (buflen<9) return MPK_INVALID_PAR;
		stgettime(st->st, &hr, &mi, &se, &st->status);
		if (st->status!=0) return st->status;
		sprintf(buf, "%02d:%02d:%02d", hr, mi, se);
		return 0;
	}
	/* Physical keys */
	stgkys(st->st, key, buf, buflen, &st->status);
	return st->status;
}

int st_getl(void *handle, const char *key, long *value)
/* Reads integer value from header */
{
	st_file *st;

	*value = 0;
	if (handle==NULL) return MPK_CLOSED_FILE;	/* Invalid handle */
	st = (st_file *) handle;
	/* Virtual keys */
	if (strcmp(key, "_width_")==0) {
		stgkyj(st->st, "Width", value, &st->status);
		return st->status;
	}
	if (strcmp(key, "_height_")==0) {
		stgkyj(st->st, "Height", value, &st->status);
		return st->status;
	}
	if (strcmp(key, "_bitpix_")==0) {
		*value = 16;
		return 0;
	}
	/* Physical keys */
	stgkyj(st->st, key, value, &st->status);
 	return st->status;
}

int st_getd(void *handle, const char *key, double *value)
/* Reads floating point value from header */
{
	st_file *st;

 	*value = 0.0;
	if (handle==NULL) return MPK_CLOSED_FILE;	/* Invalid handle */
	st = (st_file *) handle;
	/* Virtual keys */
	if (strcmp(key, "_exposure_")==0) {
		stgkyd(st->st, "Exposure", value, &st->status);
		if (st->status==0) *value = *value/100.0;
		return st->status;
	} else
	if (strcmp(key, "_temperature_")==0) {
  		stgkyd(st->st, "Temperature", value, &st->status);
  		if (st->status!=0) *value = -999.99;
		return st->status;
	} else
	if (strcmp(key, "_background_")==0) {
  		stgkyd(st->st, "Background", value, &st->status);
  		if (st->status!=0) *value = 100.0;
		return st->status;
	} else
	if (strcmp(key, "_range_")==0) {
  		stgkyd(st->st, "Range", value, &st->status);
  		if (st->status!=0) *value = 100.0;
		return st->status;
	} else {
 	    /* Locate a key in the header */
	   stgkyd(st->st, key, value, &st->status);
	   return st->status;
    }
}

int st_gkyn(void *handle, int nkey, char *key, int keylen, char *val, int vallen, char *com, int comlen)
/* Get key name and value by index. Comm is set always to "" */
{
	st_file *st;

	if (handle==NULL) return MPK_CLOSED_FILE;	/* Invalid handle */
	st = (st_file *) handle;
	return stgkyn(st->st, nkey, key, val, com, &st->status);
}

int st_g2dd(void *handle, double *ccd, int flipx, int flipy)
{
	st_file *st;

	if (handle==NULL) return MPK_CLOSED_FILE;	/* Invalid handle */
	st = (st_file *) handle;
	stg2dd(st->st, flipx, flipy, ccd, &st->status);
	return st->status;
}

int st_convhdr(void *handle, fitsfile *fits)
/* Converts header from SBIG file to FITS file */
{
  int  nx, ny, sstat=0, dstat=0, yr, mo, dy, hr, mi, se;
  long i, naxes[2], li;
  double f;
  char actdate[MAXLINE], buf[MAXLINE];
  time_t tsec;
  struct tm *actim;

  st_file *st = (st_file *) handle;
  /* Set image size */
  stghpr(st->st, &nx, &ny, &sstat);
  if (sstat!=0) return sstat;
  naxes[0]=nx; naxes[1]=ny;
  ffphps(fits,16,2,naxes,&dstat);
  if (dstat!=0) return dstat;
  /* Set date and time */
  stgetdate(st->st, &yr, &mo, &dy, &st->status);
  if (sstat!=0) return sstat;
  stgettime(st->st, &hr, &mi, &se, &st->status);
  if (sstat!=0) return sstat;
  dstat = fits_setdt(fits, yr, mo, dy, hr, mi, se, 0);
  if (dstat!=0) return dstat;
  tsec = time(&tsec);
  actim = localtime(&tsec);
  sprintf(actdate,"%04d-%02d-%02d",actim->tm_year+1900,actim->tm_mon+1,actim->tm_mday);
  ffpkys(fits,"DATE",actdate,"DATE OF CREATION",&dstat);
  if (dstat!=0) return dstat;
  sprintf(buf,"%s (%s) %s", PROGNAME, PKGNAME, VERSION);
  ffpkys(fits,"ORIGIN",buf,"",&dstat);
  ffpkyf(fits,"BSCALE",1.0,1," ",&dstat);
  ffpkyf(fits,"BZERO",BIAS,1," ",&dstat);
  stgkyj(st->st,"Exposure",&i,&sstat);
  if (sstat==0) ffpkyg(fits,"EXPTIME",1.0*i/100,2,"EXPOSURE IN SECONDS",&dstat);
  stgkyj(st->st,"Number_exposures",&i,&sstat);
  if (sstat==0) ffpkyj(fits,"SNAPSHOT",i,"NUMBER OF SNAPSHOT",&dstat);
  stgkyj(st->st,"Each_exposure",&i,&sstat);
  if (sstat==0) ffpkyg(fits,"EACHSNAP",1.0*i/100,2,"SNAPSHOT IN SECONDS",&dstat);
  stgkyd(st->st,"X_pixel_size",&f,&sstat);
  if (sstat==0) ffpkyd(fits,"PIXWIDTH",f,3,"PIXEL WIDTH IN MM",&dstat);
  stgkyd(st->st,"Y_pixel_size",&f,&sstat);
  if (sstat==0) ffpkyd(fits,"PIXHEIGH",f,3,"PIXEL HEIGHT IN MM",&dstat);
  stgkyj(st->st,"Readout_mode",&i,&sstat);
  if (sstat==0) ffpkyj(fits,"RESMODE",i,"RESOLUTION MODE",&dstat);
  stgkyj(st->st,"Exposure_state",&i,&sstat);
  if (sstat==0) {
	  li = i;
      sprintf(buf,"%lx",li);
      strup(buf);
	  ffpkys(fits,"EXPSTATE",buf,"EXPOSURE STATE (HEX)",&dstat);
  }
  stgkyd(st->st,"Temperature",&f,&sstat);
  if (sstat==0) ffpkyg(fits,"TEMPERAT",f,2,"TEMPERATURE IN DEGREES C",&dstat);
  stgkyd(st->st,"Response_factor",&f,&sstat);
  if (sstat==0) ffpkyg(fits,"RESPONSE",f,3,"CCD RESPONSE FACTOR",&dstat);
  stgkyd(st->st,"E_Gain",&f,&sstat);
  if (sstat==0) ffpkyg(fits,"EPERADU",f,2,"ELECTRONS PER ADU",&dstat);
  stgkyd(st->st,"Focal_length",&f,&sstat);
  if (sstat==0) ffpkyd(fits,"FOCALLEN",f,3,"FOCAL LENGHT IN INCHES",&dstat);
  stgkyd(st->st,"Aperture",&f,&sstat);
  if (sstat==0) ffpkyd(fits,"APERTURE",f,4,"APERTURE AREA IN SQ-INCHES",&dstat);
  stgkys(st->st,"Filter",buf,MAXLINE,&sstat);
  if (sstat==0) ffpkys(fits,"FILTER",buf,"OPTICAL FILTER NAME",&dstat);
  stgkyj(st->st,"Background",&i,&sstat);
  if (sstat==0) ffpkyj(fits,"BACKGRND",i,"BACKGROUND FOR DISPLAY",&dstat);
  stgkyj(st->st,"Range",&i,&sstat);
  if (sstat==0) ffpkyj(fits,"RANGE",i,"RANGE FOR DISPLAY",&dstat);
  stgkys(st->st,"History",buf,MAXLINE,&sstat);
  if (sstat==0) ffphis(fits,buf,&dstat);
  stgkyj(st->st,"Pedestal",&i,&sstat);
  if (sstat==0) ffpkyj(fits,"PEDESTAL",i,"PEDESTAL OFFSET",&dstat);
  stgkyj(st->st,"Track_time",&i,&sstat);
  if (sstat==0) ffpkyj(fits,"TRAKTIME",i,"TRACK PERIOD IN SECONDS",&dstat);
  stgkyj(st->st,"Sat_level",&i,&sstat);
  if (sstat==0) ffpkyj(fits,"SATURATE",i,"SATURATION LEVEL",&dstat);

  return dstat;
}

int st_convimg(void *handle, fitsfile *fits, int flipx, int flipy)
/* Converts image data from SBIG file to FITS file */
{
  double *image;
  int sstat=0, dstat=0, nx, ny;
  st_file *st;

  if (handle==NULL) return MPK_CLOSED_FILE;	/* Invalid handle */
  st = (st_file *) handle;
  /* Get image size */
  stghpr(st->st, &nx, &ny, &st->status);
  if (st->status!=0) return MPK_INVALID_HEADER;
  /* Copy image data */
  image = (double*) malloc(nx*ny*sizeof(double));
  if (image==NULL) return MPK_MEMORY;
  stg2dd(st->st,flipx,flipy,image,&sstat);
  if (sstat!=0) { free(image); return sstat; }
  ffp2dd(fits,1,nx,nx,ny,image,&dstat);
  if (dstat!=0) { free(image); return dstat; }
  /* Normal return */
  free(image);
  return 0;
}
