/**************************************************************

oesfile.c (C-Munipack project)
OES Astro read functions
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "comfun.h"
#include "oesfile.h"

#define HEADSIZE sizeof(struct __oes_header)

#define H_SIGN_SIGNED 1
#define H_SIGN_UNSIGNED 0

#define H_SUBTYPE_FIX 1
#define H_SUBTYPE_NORM 2
#define H_TYPE_BYTE 4
#define H_TYPE_FIXBYTE (H_TYPE_BYTE | H_SUBTYPE_FIX)
#define H_TYPE_WORD 8
#define H_TYPE_FIXWORD (H_TYPE_WORD | H_SUBTYPE_FIX)
#define H_TYPE_LONG 16
#define H_TYPE_FIXLONG (H_TYPE_LONG | H_SUBTYPE_FIX)
#define H_TYPE_FLOAT 32
#define H_TYPE_NFLOAT (H_TYPE_FLOAT | H_SUBTYPE_NORM)
#define H_TYPE_DOUBLE 64
#define H_TYPE_NDOUBLE (H_TYPE_DOUBLE | H_SUBTYPE_NORM)
#define H_TYPE_COMPLEX 128
#define H_TYPE_DCOMPLEX 192
#define H_TYPE_CUSTOM 256

#define H_STAT_UNDEF 0
#define H_STAT_TIME 1
#define H_STAT_IMAGE 2
#define H_STAT_COMMENT 4
#define H_STAT_DISPLAY 8
#define H_STAT_CAMERA 16

#define H_IMAG_ROTDEF 1
#define H_IMAG_PIXDEF 2
#define H_IMAG_IMGDEF 4
#define H_IMAG_LOCDEF 8
#define H_IMAG_EASTLEFT 16

#define H_DISP_HISTO 1

#define H_CAME_INTEGRATION 1
#define H_CAME_INTEST 2
#define H_CAME_TRANSEST 4
#define H_CAME_ABSTEMP 8
#define H_CAME_DARKCORR 16
#define H_CAME_FLATCORR 32
#define H_CAME_BINNING 64

#define PDAT_Y(n) ((int)(n)/16/32)
#define PDAT_M(n) ((int)(n)/32%16)
#define PDAT_D(n) ((int)(n)%32)

#define PDEG_D(n) ((int)(n)/3600)
#define PDEG_M(n) ((int)(n)/60%60)
#define PDEG_S(n) ((int)(n)%60)

#define PTIM_H(n) ((int)(n)/3600)
#define PTIM_M(n) ((int)(n)/60%60)
#define PTIM_S(n) ((int)(n)%60)

struct __oes_header {
	/* Contents */
	long dx, dy;         /* Picture dimensions */
	long pxlen;          /* Bytes per pixel */
	long sign, type;     /* Signed/unsigned, data type */
	long status;         /* Which parts are present? */
	/* Time */
	float ymd, hms;      /* Date and time of observation */
	float stime, jd;     /* Star time and julian date */
	/* Image */
	long seqnum, frnum;  /* Number of sequence, frame */
	long istatus;        /* Image status */
	float irot;          /* Image rotation */
	float pxwidth;       /* Pixel width */
	long pxrationdividend, pxrationdivider; /* Aspect ratio */
	float ra, dec;       /* R.A., DEC. */
	float lon, lat;      /* Longitude, latitude */
	/* Display */
	long dstatus;        /* Display status */
	float dmax, dmin;    /* Display params (White, Black) */
	long xmax, ymax;
	long xmin, ymin;     /* histogram display params */
	/* Camera */
	long cstatus;        /* Camera status */
	long camera;         /*  */
	float exposure;      /* duration of exposure */
	float transfer;      /* duration of transfer */
	float temp;          /* temperature */
	/* Reserved */
	long res[27];        /* Future expansion */
	float fixoffset;     /* Minimum of fixed numbers */
	float fixdynamics;   /* Fixed numbers dynamics */
	unsigned long magic1, magic2; /* Magic numbers (endian type detection) */
	/* Comments */
	char comment[4][64]; /* Location, telescope */
};

struct __oesfile {
	FILE *f;
	struct __oes_header head;
	char *ccd;
};

typedef struct __oesfile __oesfile;

const char *keys[42] = {
	"OES Astro file", "dX", "dY", "PxLen", "Sign", "Type", "Status", "YeMoDa", "HoMiSe", "StarTime", "JulDat", "SeqNum",
	"FrameNum", "ImagStat", "ImagRot", "PxWidth", "PxRtDiv", "PxRtDis", "Rektas", "Deklin", "Longitude", "Latitude",
	"DispStat", "Maximum", "Minimum", "Xmax", "Ymax", "Xmin", "Ymin", "CamStat", "Camera", "Integration", "Transfer",
	"Temperature", "FixOffset", "FixDynamic", "Magic1", "Magic2", "Comment1", "Comment2", "Comment3", "Comment4"
};

const char *coms[42] = {
	"file format name", "image width in pixels", "image height in pixels", "number of bytes per pixel", NULL,
	NULL, "filled parts of header", "date of observation", "time of observation", "star time of observation",
	"julian date of observation", "sequence number", "frame number in the sequence", "filled parts in the header",
	"rotation of the field", "pixel width in arc seconds", "pixel ratio (nominator)", "pixel ratio (denominator)",
	"object's right ascension", "object's declination", "longitude of observation place", "latitude of observation place",
	"display status", "white point in ADU", "black point in ADU", NULL, NULL, NULL, NULL, "camera status", "camera id",
	"integration duration", "transfer duration", "temperature", NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};

int oesopen(oesfile **fs, const char *filename)
// Opens a file and reads the header
{
  __oesfile *oes;
  size_t headlen;

  /* File structure allocation */
  oes = (__oesfile *)malloc(sizeof(__oesfile));
  if (oes==NULL) return MPK_MEMORY;        /* Insufficient memory */
  memset(oes, 0, sizeof(__oesfile));

  /* Open file */
  oes->f = fopen(filename,"rb");
  if (oes->f==NULL) {
	  oesclos(oes);
	  return MPK_CANNOT_OPEN_SRC;          /* Cannot open file */
  }

  /* Read header of the file */
  headlen = fread(&oes->head,1,HEADSIZE,oes->f);
  if ((headlen!=HEADSIZE)
      || ((oes->head.dx<=0)||(oes->head.dx>0x4000)||(oes->head.dy<=0)||(oes->head.dy>0x4000))
      || (((oes->head.magic1!=H_MAG1_OWN)||(oes->head.magic2!=H_MAG2_OWN))
        &&((oes->head.magic1!=H_MAG1_INV)||(oes->head.magic2!=H_MAG2_INV)))) {
	  oesclos(oes);
      return MPK_INVALID_HEADER;
  }

  /* Normal return */
  *fs = oes;
  return 0;
}

void oesclos(oesfile *fs)
/* Free allocated memory */
{
  __oesfile *oes = (__oesfile *)fs;
  if (oes) {
	  if (oes->f) fclose(oes->f);
	  if (oes->ccd) free(oes->ccd);
  	  free(oes);
  }
}

int oesghpr(oesfile *fs, int *nx, int *ny)
{
 __oesfile *oes = (__oesfile *)fs;

 if (oes==NULL) return MPK_CLOSED_FILE;
 *nx = oes->head.dx;
 *ny = oes->head.dy;
 return 0;
}

int oesg2dd(oesfile *fs, int flipx, int flipy, double saturate, double *buf)
{
 __oesfile *oes = (__oesfile *)fs;
 int nx, ny, i, j, ii, jj;
 size_t datalen;

 if (oes==NULL) return MPK_CLOSED_FILE;

 nx = oes->head.dx;
 ny = oes->head.dy;
 if ((nx<=0)||(nx>0x4000)||(ny<=0)||(ny>0x4000))
     return MPK_INVALID_HEADER;

 if (oes->ccd==NULL) {
     /* Allocate the buffer */
	 // POZOR!!! Tady se to musi spravit, pokud se objevi jine datovy typy!
     oes->ccd = (char *)malloc(nx*ny*sizeof(unsigned int));
     if (oes->ccd==NULL) return MPK_MEMORY;        /* Insufficient memory */
     /* Read data into the buffer */
     fseek(oes->f, HEADSIZE, SEEK_SET);
     datalen = fread(oes->ccd,1,4*nx*ny,oes->f);
     if ((int)datalen!=4*nx*ny) return MPK_INVALID_DATA;
 }

 if (oes->head.type==H_TYPE_NFLOAT) {
	 /* Convert data from buffer to array of floats */
     for (j=0;j<ny;j++) {
	     for (i=0;i<nx;i++) {
		     if (flipx) ii = nx-i-1; else ii = i;
			 if (flipy) jj = ny-j-1; else jj = j;
			 buf[j*nx+i] = ((float *)(oes->ccd))[jj*nx+ii]*saturate;
			 if (buf[j*nx+i]<=0.0) buf[j*nx+i]=0.0;
			 if (buf[j*nx+i]>=(saturate-1)) buf[j*nx+i]=saturate-1;
		 }
	 }
 } else
     return MPK_INVALID_DATA;                    /* Unsupported data format */
 return 0;
}

int oesgexp(oesfile *fs, double *exp)
// Get exposure length in seconds
{
 __oesfile *oes = (__oesfile *)fs;
 if (oes==NULL) return MPK_CLOSED_FILE;

 *exp = PTIM_H(oes->head.exposure)*3600.0+PTIM_M(oes->head.exposure)*60.0+PTIM_S(oes->head.exposure)*1.0;
 if ((*exp<=0)||(*exp>=99999.9)) {
	 *exp = -1.0;
     return MPK_INVALID_HEADER;
 } else
	 return 0;
}

int oesgtem(oesfile *fs, double *temp)
// Get exposure length in seconds
{
 __oesfile *oes = (__oesfile *)fs;
 if (oes==NULL) return MPK_CLOSED_FILE;

 *temp = oes->head.temp;
 if ((*temp<-99.9)||(*temp>99.9)) {
     *temp = 999.9;                            /* This is invalid temperature */
     return MPK_INVALID_HEADER;
 } else
	 return 0;
}

int oesgdat(oesfile *fs, int *yr, int *mon, int *day, int *hr, int *min, int *sec)
// Reads date and time
{
 __oesfile *oes = (__oesfile *)fs;
 if (oes==NULL) return MPK_CLOSED_FILE;

 *yr = PDAT_Y(oes->head.ymd); *mon = PDAT_M(oes->head.ymd); *day = PDAT_D(oes->head.ymd);
 *hr = PTIM_H(oes->head.hms); *min = PTIM_M(oes->head.hms), *sec = PTIM_S(oes->head.hms);
 if ((*yr>1900)&&(*yr<=9999)&&(*mon>=1)&&(*mon<=12)&&(*day>=1)&&(*day<=31)
        &&(*hr>=0)&&(*hr<=23)&&(*min>=0)&&(*min<=59)&&(*sec>=0)&&(*sec<=59)) {
	 return 0;
 } else {
	 *yr = *mon = *day = *hr = *min = *sec = 0;
     return MPK_INVALID_DATE;
 }
}

int oesgkyj(oesfile *fs, int nkey, long *value)
// Returns value from the header as long
{
 __oesfile *oes = (__oesfile *)fs;
 if (oes==NULL) return MPK_CLOSED_FILE;
 if ((nkey<0)||(nkey>=(int)(HEADSIZE/4))) return MPK_KEY_NOT_FOUND;

 *value = ((long *)&(oes->head))[nkey];
 return 0;
}

int oesgkyd(oesfile *fs, int nkey, double *value)
// Returns value from the header as double
{
 __oesfile *oes = (__oesfile *)fs;
 if (oes==NULL) return MPK_CLOSED_FILE;
 if ((nkey<0)||(nkey>=(int)(HEADSIZE/4))) return MPK_KEY_NOT_FOUND;

 *value = ((float*)&(oes->head))[nkey];
 return 0;
}

int oesgkys(oesfile *fs, int nkey, char *buf, int buflen)
// Returns value from the header as string
{
 __oesfile *oes = (__oesfile *)fs;
 if (oes==NULL) return MPK_CLOSED_FILE;
 if ((nkey<64)||(nkey>=128)||(nkey%16!=0)) return MPK_KEY_NOT_FOUND;

 if (buf) {
	 int len = (buflen<65 ? buflen-1 : 64);
	 strncpy(buf, oes->head.comment[(nkey-64)/16], len);
	 buf[len] = 0;
 }

 return 0;
}

int oesgkyn(oesfile *fs, int nkey, char *key, char *val, char *com)
     // Sequential read of parameters
{
	long i;
	int y, m, d, h, n;
	double f, s;
	__oesfile *oes = (__oesfile *)fs;

	if (oes==NULL || nkey<0 || nkey>=42)
		return MPK_INVALID_PAR;

	if (key) {
		if (keys[nkey]) strcpy(key, keys[nkey]);
		else strcpy(key, "");
	}
	if (com) {
		if (coms[nkey]) strcpy(com, coms[nkey]);
		else strcpy(com, "");
	}
	if (val) {
		strcpy(val, "");
	}

	switch(nkey)
	{
	case 1:	case 2:	case 3:	case 6:	case 11: case 12: case 13: case 16:
	case 17: case 22: case 25: case 26: case 27: case 28: case 29: case 30:
		oesgkyj(oes,nkey-1,&i);
		if (val) sprintf(val, "%ld", i);
		break;

	case 4:
		oesgkyj(oes,3,&i);
		if (val) sprintf(val, "%ld", i);
		if (com) switch (i) {
			case 0: strcpy(com, "unsigned data type"); break;
			case 1: strcpy(com, "signed data type"); break;
			default: strcpy(com, "invalid value");
		}
		break;

	case 5:
		oesgkyj(oes,4,&i);
		if (val) sprintf(val, "%ld", i);
		if (com) {
			switch (i&0xFFFC) {
				case 4:   strcpy(com, "8-bit integer"); break;
				case 8:   strcpy(com, "16-bit integer"); break;
				case 16:  strcpy(com, "32-bit integer"); break;
				case 32:  strcpy(com, "32-bit float"); break;
				case 64:  strcpy(com, "64-bit float"); break;
				case 128: strcpy(com, "32-bit complex"); break;
				case 192: strcpy(com, "64-bit complex"); break;
				case 256: strcpy(com, "custom format"); break;
				default:  strcpy(com, "invalid value");
			}
			if ((i&1)!=0)
				strcat(com, " with fixed offset");
			if ((i&2)!=0)
				strcat(com, " normalized to 0..1");
		}
		break;

	case 7:
		oesgkyd(oes,6,&f);
		y=(int)(f/(16.0*32.0)); m=(int)((f-y*16.0*32.0)/32.0); d=(int)((f-y*16.0*32.0-m*32.0));
		if (val) sprintf(val, "%04d-%02d-%02d", y, m, d);
		break;

	case 8: case 9: case 14: case 18: case 19: case 20: case 21: case 31: case 32:
		oesgkyd(oes,nkey-1,&f);
		h=(int)(f/(60.0*60.0)); n=(int)((f-h*60.0*60.0)/60.0); s=(f-h*60.0*60.0-n*60.0);
		if (val) sprintf(val, "%02d:%02d:%04.1f", h, n, s);
		break;

	case 10:
		oesgkyd(oes,9,&f);
		if (val) sprintf(val, "%.6f", f);
		break;

	case 15: case 23: case 24: case 33:
		oesgkyd(oes,nkey-1,&f);
		if (val) sprintf(val, "%.2f", f);
		break;

	case 34: case 35:
		oesgkyd(oes,nkey+26,&f);
		if (val) sprintf(val, "%.2f", f);
		break;

	case 36:
		oesgkyj(oes,62,&i);
		sprintf(val, "%08lX", i);
		switch (i) {
			case 0x012AEE94: strcpy(com, "Own magic number"); break;
			case 0x94EE2A01: strcpy(com, "Swapped magic number"); break;
			default: strcpy(com, "Invalid magic number");
		}
		break;

	case 37:
		oesgkyj(oes,63,&i);
		sprintf(val, "%08lX", i);
		switch (i) {
			case 0x012F3462: strcpy(com, "Own magic number"); break;
			case 0x62342F01: strcpy(com, "Swapped magic number"); break;
			default: strcpy(com, "Invalid magic number");
		}
		break;

	case 38: case 39: case 40: case 41:
		oesgkys(oes,(nkey-38)*16+64, com, 79);
		break;

	default:
		return MPK_KEY_NOT_FOUND;
	}
	return 0;
}
