/**************************************************************

darkbat.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fitsio.h>

#ifdef _WIN32
#include <io.h>
#endif

#include "comfun.h"
#include "mdark.h"

void PrintHelp()
{
 printf("MEANDARK version %s - utility for the computing master-dark fields.\n",VERSION);
 printf(PROJECTLINE);
 printf("\nUsage:meandark [options] [@dirfile] [out=file] [image1] ...\n\n");
 printf("parameters:\n");
 printf("     @dirfile=  file with input image names\n");
 printf("                (for only '@', read image names from standart input)\n");
 printf("     par=file   file with parameters (default = %s)\n",OPTFILE);
 printf("     bias=file  make scalable dark, file with bias frame\n");
 printf("     out=       output file name (default = %s)\n",MDARK_MASKNAME);
 printf("     image1...  name(s) of simple input image(s)\n");
 printf("options:\n");
 printf("     -p<name>   parameter file (equivalent with 'par=')\n");
 printf("     -o<mask>   output file name (equivalent with 'mask=')\n");
 printf("     -b<file>   bias frame file (equivalent with 'bias=')\n");
 printf("     -q         quiet mode\n");
 printf("     -h         give this help\n");
 printf("     -v         prints software version\n");
 printf("     -l         display software license\n");
 printf("     -g         creates the example of the parameter and aperture files\n");
 printf(WARRANTYLINE);
 printf(HOMEPAGELINE);
 printf(BUGREPORTS);
}

void report(int status, const char *file)
{
 if (cmpack_debug) 
 	fprintf(stderr, "The following error(s) has been occurred: %d\n",status);
 if (status<1000) 
 	ffrprt(stderr, status);
 else  
 	PrintError(MDARK_PROGNAME, file, 0, status);
}

void error(int status, const char *file)
{
 report(status, file);
 exit(status);
}

void CreatePar()
/* Creates a sample parameter file */
{
 FILE *fp;

 if ((fp=fopen(SOPFILE,"w"))!=NULL) {
   printf("Writing '%s' ...\n\n",SOPFILE);
   fprintf(fp,"BITPIX = 0         # Output image data format (0=Auto)\n");
   fclose(fp);
   printf("'%s' is an example of the parameter file. Rename it to '%s' and edit the parameters.\n\n",
       SOPFILE, OPTFILE);
 }
}

int main(int argc, char **argv)
{
    int stdinput = FALSE, i, res, level = 0;
    char line[MAXLINE], buf1[MAXLINE];
    char *out = MDARK_MASKNAME, *nampar = NULL, *namsci = "", *nambias = "", *par;
	FILE *f;

	/* Zpracovani parametru prikazove radky */
	for (i=1;i<argc;i++) {
	   par = argv[i];
	   /* Prints help */
	   if ((strcmp(par,"-h")==0)||(strcmp(par,"--help")==0)) {
		   PrintHelp();
          return 0;
       }
	   /* Prints version */
	   if ((strcmp(par,"-v")==0)||(strcmp(par,"--version")==0)) {
		  PrintVersion(MDARK_PROGNAME);
          return 0;
       }
       /* Prints license */
       if ((strcmp(par,"-l")==0)||(strcmp(par,"--license")==0)) {
          PrintLicense();
          return 0;
       }
       /* Generate sample parameter file */
       if (strcmp(argv[i],"-g")==0) {
           CreatePar();
           return 0;
       }
       /* Flag for input from stdin or dirfile */
       if (strchr(par,'@')==par) {
	       namsci = malloc(strlen(par));
	       strcpy(namsci,&par[1]);
           stdinput = strlen(namsci)==0;
       }
       /* Parameter file */
       if (strstr(par,"par=")==par) {
	       nampar = malloc(strlen(par)-3);
	       strcpy(nampar,&par[4]);
       }
       if (strstr(argv[i],"-p")==par) {
	       nampar = malloc(strlen(par)-1);
	       strcpy(nampar,&par[2]);
       }
       /* File with bias frame */
       if (strstr(par,"bias=")==par) {
	       nambias = malloc(strlen(par)-4);
	       strcpy(nambias,&par[5]);
       }
       if (strstr(par,"-b")==par) {
	       nambias = malloc(strlen(par)-1);
	       strcpy(nambias,&par[2]);
       }
       /* Name of output file */
       if (strstr(par,"out=")==par) {
	       out = malloc(strlen(par)-3);
	       strcpy(out,&par[4]);
       }
       if (strstr(par,"-o")==par) {
	       out = malloc(strlen(par)-1);
	       strcpy(out,&par[2]);
       }
       /* Quiet/verbose flags */
       if (strcmp(par,"-q")==0 || strcmp(par,"--quiet")==0 || strcmp(par,"--silent")==0) {
           level = MPK_LEVEL_QUIET;
       }
       if (strcmp(par,"--verbose")==0 || strcmp(par,"--debug")==0) {
           level = MPK_LEVEL_DEBUG;
       }
  }

  cmpack_setoutput(PrintMessage, level);
  PrintPrologue(MDARK_PROGNAME);
  res=mdark_init(nampar, strlen(nambias)>0, nambias);
  if (res!=0) error(res, NULL);

  /* image names reads from command line */
  for (i=1;i<argc;i++) {
	  par = argv[i];
      if ((strchr(par,'=')==0)&&(strchr(par,'@')!=par)&&(strchr(par,'-')!=par)) {
#ifdef _WIN32
	    intptr_t fh;
		struct _finddata_t fd;
		if ((fh=_findfirst(par, &fd))!=-1) {
			do {
				if ((fd.attrib & _A_SUBDIR)==0)  {
          			res=mdark_read(fd.name);
          			if (res!=0) report(res, fd.name);
          		}
          	} while (_findnext(fh, &fd)==0);
			_findclose(fh);
      	 } else {
	      	 report(MPK_CANNOT_OPEN_SRC, par);
      	 }
#else
		res=mdark_read(par);
		if (res!=0) report(res, par);
#endif
      }
  }

  /* image names reads from dirfile */
  if (strlen(namsci)>0) {
	  f = fopen(namsci,"r");
	  if (f==NULL) error(MPK_NO_DIRFILE, namsci);
      while(fgets(buf1,MAXLINE,f)!=NULL) {
          sscanf(buf1,"%s",line);
          if (strcmp(line,"")) {
	          res=mdark_read(line);
     	      if (res) report(res, line);
          }
      }
  }

  /* Image names reads from stdin */
  if (stdinput) {
      while(fgets(buf1,MAXLINE,stdin)!=NULL) {
          sscanf(buf1,"%s",line);
          if (strcmp(line,"")) {
	          res=mdark_read(line);
     	      if (res) report(res, line);
          }
      }
  }

  /* Write output */
  res=mdark_write(out);
  if (res!=0) report(res, out);

  res=mdark_clean();
  if (res!=0) error(res, NULL);

  return 0;
}
