/**************************************************************

vfind.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "comfun.h"
#include "trajd.h"
#include "inifile.h"
#include "mfind.h"

int        mf_aper;           // Aperture number
int        mf_thr;            // Threshold in percents
int        mf_refstar;        // Index of reference star (-1 = Autodetection)
int        mf_nimages;        // No. of images
int        mf_nstars;         // No. of stars
double     mf_sigma;          // Threshold 2

#define avg(x,z)       ((x)/(z))
#define stdev(x,y,z)   sqrt(((y)-((x)*(x)/(z)))/((z)-1))
#define MAG(i,j)       mf_data[((i)*mf_nstars+(j))*2]
#define ERR(i,j)       mf_data[((i)*mf_nstars+(j))*2+1]

struct MagInfo
{
	int j;                        /* Poradi hvezdy */
	double mag, err;               /* Magnituda, chyba */
	struct MagInfo *next;         /* Dalsi mereni nebo NULL */
};

struct ImageInfo
{
	double jd;                    /* Julian date */
    struct MagInfo  *first, *last;/* Mereni hvezd */
    struct ImageInfo *next;       /* dalsi snimek nebo NULL */
};

struct StarInfo
{
	int    valid, npoints;
	double minmag, maxmag;
	double mean, stdev;
};

double     *mf_data;           // Array of mag,err data
struct StarInfo  *mf_info;           // Array of star info
struct ImageInfo *mf_first;          // Stack of images (pointer to first image)
struct ImageInfo *mf_last;           // Stack of images (pointer to last image)

int mfind_init(const char *optfile, int ap, int thr)
/* Initializes parameters and opens output file */
{
  inifile *f = NULL;
  char msg[MAXLINE];

  // Init parameters
  mf_nimages = 0;
  mf_nstars  = 0;
  mf_refstar = -1;
  mf_sigma   = 3.0;
  mf_data    = NULL;
  mf_info    = NULL;
  mf_first   = NULL;
  mf_last    = NULL;

  // Read parameter file
  f = iniopen(optfile, OPTFILE);
  if (f==NULL) {
	if (optfile) sprintf(msg,"Cannot read parameter file '%s'", optfile);
		    else sprintf(msg,"Cannot read parameter file '%s'", OPTFILE);
    PrintOut(0, msg);
  	return MPK_CANNOT_OPEN_PAR;
  }
  mf_aper = inigkyl(f, "APERTURE",  DEFAPERTURE);
  mf_thr  = inigkyl(f, "THRESHOLD", DEFTHRESHOLD);
  iniclos(f);

  // Function parameter overwrites the value in parameter file
  if (ap>0)   mf_aper = ap;
  if (thr>=0) mf_thr  = thr;

  // Check parameters
  if (mf_aper<=0) {
      sprintf(msg, "Invalid number of aperture (%d)", mf_aper);
	  PrintOut(0, msg);
	  return MPK_INVALID_PAR;
  }
  if ((mf_thr<0)||(mf_thr>100)) {
      sprintf(msg, "Invalid value of threshold (%d)", mf_thr);
      PrintOut(0, msg);
	  return MPK_INVALID_PAR;
  }

  if (ap>0) {
      sprintf(msg, "Aperture: %d", mf_aper);
      PrintOut(0, msg);
  }
  if (thr>=0) {
      sprintf(msg, "Threshold: %d", mf_thr);
      PrintOut(0, msg);
  }

  // Normal return
  return 0;
}

int mfind_readsrt(const char *infile)
/* Reads magnitudes and error of the selected stars and writes them to output file */
{
  FILE *f;
  char buf[MAXLINE], *end, *aux;
  double mag, err, jd;
  int j,delta,nl=0;
  struct ImageInfo *ptr;
  struct MagInfo  *dta;

  /* Prints name of input file */
  PrintOut(0, infile);

  /* Read input file */
  f = fopen(infile,"r");
  if (f==NULL) return MPK_CANNOT_OPEN_SRC;

  /* Reads header information */
  fgets(buf,MAXLINE,f);
  if ((strstr(buf," NL")==buf)&&(strstr(buf,"JD")!=NULL)) {
      fgets(buf,MAXLINE,f);
      nl = atol(buf);
      jd = atof(buf+62);
  }
  if (nl!=2) {
	  fclose(f);
	  return MPK_UNKNOWN_FORMAT;
  }

  /* Create a new image record */
  ptr = (struct ImageInfo *) malloc(sizeof(struct ImageInfo));
  if (!ptr) {
	  fclose(f);
	  return MPK_MEMORY;
  }
  ptr->jd    = jd;
  ptr->next  = NULL;
  ptr->first = NULL;
  ptr->last  = NULL;
  if (mf_last) { mf_last->next = ptr; mf_last = ptr; }
          else { mf_last = ptr; mf_first = ptr; }
  mf_nimages++;

  /* Read magnitudes and errors of stars */
  j = 2; mag = 99.999; err = 9.9999;
  while (fgets(buf,MAXLINE-1,f)) {
      j++;
      if ((j%3)==1) {
	      mag = 99.999;
          delta=25+9*(mf_aper-1);
	      if ((int)strlen(buf)>delta) {
	      	 aux = buf + delta;
          	 mag = strtod(aux,&end);
             if (end==aux) mag = 99.999;
          }
      }
      if ((j%3)==2) {
	      err = 9.9999;
          delta=26+9*(mf_aper-1);
          if ((int)strlen(buf)>delta) {
              aux = buf + delta;
          	  err = strtod(aux,&end);
          	  if (end==aux) err = 9.9999;
          }
	      //fprintf(stdout," %d %7.3f +- %7.3f\n",j,mag,err);
	      if ((mag<99)&&(err<9)) {
  	         dta = (struct MagInfo *) malloc(sizeof(struct MagInfo));
  	         if (ptr->last) { ptr->last->next = dta; ptr->last = dta; }
	                   else { ptr->last = dta; ptr->first = dta; }
             dta->j = (j/3);
	         dta->err = err;
	         dta->mag = mag;
	         dta->next = NULL;
	         if (dta->j > mf_nstars) mf_nstars = dta->j;
	      }
      }
  }
  fclose(f);

  /* Normal return */
  return 0;
}

int mfind_readall(const char *infile)
/* Reads magnitudes and errors */
{
  FILE *f;
  char buf[MAXLONGLINE], *end, *aux, msg[MAXLINE];
  double mag, err, jd;
  int j;
  struct ImageInfo *ptr;
  struct MagInfo  *dta;

  /* Prints name of input file */
  sprintf(msg,"Source file (all-stars format):     %s",infile);
  PrintOut(0, msg);

  /* Read input file */
  f = fopen(infile,"r");
  if (f==NULL) return MPK_CANNOT_OPEN_SRC;

  /* Reads header information */
  while (fgets(buf,MAXLONGLINE-1,f)) {

      // Strip comments and empty lines
      if ((buf[0]=='#')||(buf[0]=='\0')) continue;

      // Read JD value
      jd = strtod(buf, &end);
      if (end==buf) continue;
      aux = end;

      // Create a new image record
      ptr = (struct ImageInfo *) malloc(sizeof(struct ImageInfo));
      if (!ptr) {
	     fclose(f);
	     return MPK_MEMORY;
      }
      ptr->jd    = jd;
      ptr->next  = NULL;
      ptr->first = NULL;
      ptr->last  = NULL;
      if (mf_last) { mf_last->next = ptr; mf_last = ptr; }
              else { mf_last = ptr; mf_first = ptr; }
      mf_nimages++;

      // Read magnitudes and errors of stars
      aux = end;
      j = 1;
      while (aux[0]!=0) {
          mag = 99.999;
          err =  9.999;

          mag = strtod(aux, &end);
          if (end==aux) mag = 99.999;
          aux = end;
      	  err = strtod(aux, &end);
          if (end==aux) err = 9.9999;
          aux = end;

	      if ((mag<99)&&(err<9)) {
  	         dta = (struct MagInfo *) malloc(sizeof(struct MagInfo));
  	         if (ptr->last) { ptr->last->next = dta; ptr->last = dta; }
	                   else { ptr->last = dta; ptr->first = dta; }
             dta->j    = j;
	         dta->err  = err;
	         dta->mag  = mag;
	         dta->next = NULL;
  	         //fprintf(stdout,"%d %d %4.3f %4.3f\n", mf_nimages, j, mag, err);
	      }
	      while ((aux[0]!=0)&&(aux[0]!='-')&&((aux[0]<'0')||(aux[0]>'9'))) aux++;
          j++;
      }
      if (j-1>mf_nstars) mf_nstars=j-1;
  }
  fclose(f);

  /* Normal return */
  return 0;
}

int mfind_compute(int ref)
{
  struct ImageInfo *ptr;
  struct MagInfo  *dta;
  char msg[MAXLINE];
  int npoints, p0, i, j, rj, r, rmax, nreferr, nbadpoints, nerrors, nwritten, nmeasur, nrequired;
  double sum, sumsq, q, q0, diff, x, y, *rs_mag;
  int rs_index[RSTARS], *rs_valid;

  PrintOut(0, "---------------------------------------------");
  if ((mf_nimages<2)||(mf_nstars<2))
	  return MFI_TOO_FEW_DATA;

  /* Prepare 'data' and 'info' arrays */
  if (mf_data) mf_data = (double *) malloc(mf_nimages*mf_nstars*2*sizeof(double));
          else mf_data = (double *) realloc(mf_data,mf_nimages*mf_nstars*2*sizeof(double));
  if (mf_info) mf_info = (struct StarInfo *) malloc(mf_nstars*sizeof(struct StarInfo));
          else mf_info = (struct StarInfo *) realloc(mf_info,mf_nstars*sizeof(struct StarInfo));

  ptr = mf_first;
  i = 0;
  while (ptr) {
	  dta = ptr->first;
	  j = 0;
	  while (dta) {
		  while (j<dta->j-1) { MAG(i,j)=99.999; ERR(i,j)=9.9999; j++; }
		  MAG(i,j) = dta->mag;
		  ERR(i,j) = dta->err;
		  dta = dta->next;
		  j++;
	  }
	  while (j<mf_nstars) { MAG(i,j)=99.999; ERR(i,j)=9.9999; j++; }
	  ptr = ptr->next;
	  i++;
  }

/* finding averages and standard deviations of measurement errors ->
   maximal allowed measurement error for each star */
  for (j=0; j<mf_nstars; j++) {
      npoints = 0;
	  for (i=0; i<mf_nimages; i++) {
		  if (MAG(i,j)<99) npoints++;
	  }
	  mf_info[j].npoints = npoints;
	  mf_info[j].valid = (npoints>=2);
  }

  /* Select reference star */
  if (ref>0) {
	  rj = ref-1;
  } else {
      rj = -1;
      rs_mag = (double*) malloc(mf_nimages*sizeof(double));
      rs_valid = (int*) malloc(mf_nimages*sizeof(int));
      // Select set of brightest stars
	  for (j=0, r=0; j<mf_nstars && r<RSTARS; j++) {
         if (mf_info[j].valid) {
	         rs_index[r++] = j;
         }
      }
      rmax = r;
      // Compute data for one artifitial comparison star
      for (i=0; i<mf_nimages; i++) {
	      // Compute average brightness
	      sum = 0.0; npoints = 0;
	      for (r=0; r<rmax; r++) {
		      j = rs_index[r];
  		      if (MAG(i,j)<99) {
	  		      sum += pow(10, -0.4*MAG(i,j));
	  		      npoints++;
  		      }
		  }
		  // All stars must have valid data
		  if (npoints==rmax) {
		      rs_mag[i] = -2.5*log10(sum/npoints);
		      rs_valid[i] = 1;
	      } else {
		      rs_mag[i] = 99.9;
		      rs_valid[i] = 0;
	      }
      }
      // Select one star, which has minimal std. dev. relative to the art. comp.
      q0 = 99.9; p0 = 0;
      for (r=0; r<rmax; r++) {
	      j = rs_index[r];
	      // Compute std.dev. of magnitudes relative to art. comp. star
          sum = sumsq = 0.0; npoints = 0;
          for (i=0; i<mf_nimages; i++) {
	          if (MAG(i,j)<99 && rs_valid[i]) {
                  diff   = MAG(i,j) - rs_mag[i];
                  sum   += diff;
  	              sumsq += diff * diff;
	              npoints++;
	          }    
	      }
	      // Find minimum of std.dev.
	      if (npoints>=2 && npoints>=p0) {
	          q = stdev(sum, sumsq, npoints);
	          if (npoints>p0 || q<q0) { q0 = q; p0 = npoints; rj = j; }
          }
      }
      // If failed, use star with most valid measurements 
      if (rj<0) {
	  	q0 = 0;
	  	for (j=0; j<mf_nstars; j++) {
        	if (mf_info[j].valid) {
	        	q = mf_info[j].npoints;
 	         	if (q>q0) { q0 = q; rj = j; }
 	     	}
	     }
      }      
      free(rs_mag);
      free(rs_valid);
  }
  if ((rj<0)||(rj>=mf_nstars)) return MFI_REF_NOT_FOUND;

  mf_refstar = rj+1;
  nreferr = mf_nimages - mf_info[rj].npoints;
  nrequired = (int)(0.01*mf_thr*mf_info[rj].npoints);
  if (nrequired<2) nrequired = 2;

  /* computing average and std. dev. for each star - first iteration */
  for (j=0; j<mf_nstars; j++) {
      mf_info[j].mean   = 99.9999;
	  mf_info[j].stdev  = 9.9999;
      mf_info[j].minmag = 0.0;
      mf_info[j].maxmag = 0.0;
	  if (mf_info[j].valid) {
	      sum = sumsq = 0.0; npoints = 0;
		  for (i=0; i<mf_nimages; i++) {
			  if (MAG(i,j)<99 && MAG(i,rj)<99) {
	              diff   = MAG(i,j) - MAG(i,rj);
			      sum   += diff;
			      sumsq += diff * diff;
			      npoints++;
		      }
		  }
		  if (npoints>=2 && npoints>=nrequired) {
		      x = avg(sum,npoints);
		      y = stdev(sum,sumsq,npoints);
 	          mf_info[j].minmag  = x - mf_sigma*y;
	          mf_info[j].maxmag  = x + mf_sigma*y;
	      } else {
		      mf_info[j].valid = 0;
	      }
	  	  //printf(" %5d %5d %8.3f %8.3f %8.3f %8.3f\n",j,info[j].npoints,info[j].sum,info[j].sumsq,info[j].minmag,info[j].maxmag);
  	  }
  }

  nwritten = nbadpoints = nmeasur = 0;
  for (j=0;j<mf_nstars;j++) {
	  if (mf_info[j].valid) {
	      sum = sumsq = 0.0; npoints = 0; nerrors = 0;
		  for (i=0;i<mf_nimages;i++) {
			  if (MAG(i,j)<99) {
				  nmeasur++;
				  if (MAG(i,rj)<99) {
	  			  	  diff = MAG(i,j) - MAG(i,rj);
			      	  if (diff>=mf_info[j].maxmag || diff<=mf_info[j].minmag) {
				          nerrors++;
			          } else {
	                      sum   += diff;
	                      sumsq += diff*diff;
	                      npoints++;
	                  }
	              }
	          }
	      }
	      mf_info[j].npoints = npoints;
	      if (j!=rj && npoints>=nrequired && npoints>=2) {
	          mf_info[j].mean  = avg(sum,npoints);
	          mf_info[j].stdev = stdev(sum,sumsq,npoints);
	          nwritten++;
	          nbadpoints += nerrors;
	      } else {
      		  mf_info[j].mean   = 99.9999;
	  		  mf_info[j].stdev  = 9.9999;
		      mf_info[j].valid  = 0;
	      }
      }
  }

  /* Print messages */
  sprintf(msg,"Number of analyzed stars:           %d (+ reference star #%d)",mf_nstars-1,rj+1);
  PrintOut(0, msg);
  sprintf(msg,"Number of source files:             %d",mf_nimages);
  PrintOut(0, msg);
  sprintf(msg,"Total number of data points:        %d",nmeasur);
  PrintOut(0, msg);
  sprintf(msg,"Reference star used:                %d",mf_refstar);
  PrintOut(0, msg);
  sprintf(msg,"Reference star errors:              %d (%.0f%%)",nreferr,100.0*nreferr/nmeasur);
  PrintOut(0, msg);
  sprintf(msg,"Number of good points required:     %d",nrequired);
  PrintOut(0, msg);
  sprintf(msg,"Number of stars written to output:  %d (%.0f%%)",nwritten,100.0*nwritten/(mf_nstars-1));
  PrintOut(0, msg);
  sprintf(msg,"Number of points ruled out:         %d (%.0f%%)",nbadpoints,100.0*nbadpoints/nmeasur);
  PrintOut(0, msg);

  /* Normal return */
  return 0;
}

int mfind_getpar(int *nstars, int *nimages, int *ref)
{
   if ((!mf_data)||(!mf_info)) return MPK_NO_DATA;

   *nstars  = mf_nstars;
   *nimages = mf_nimages;
   *ref     = mf_refstar;

   return 0;
}

int mfind_get(double *mag, double *stdev)
{
   	int j;

   	if ((!mf_data)||(!mf_info)) 
   		return MPK_NO_DATA;
   	if ((!mag)||(!stdev)) 
   		return MPK_INVALID_PAR;

   	for (j=0;j<mf_nstars;j++) {
		if (mf_info[j].valid) {
			mag[j]   = mf_info[j].mean;
	   		stdev[j] = mf_info[j].stdev;
   		} else {
			mag[j]   = 999.99;
	   		stdev[j] = 99.999;
   		}
	}

   return 0;
}

int mfind_getdata(int star, double *jd, double *mag, double *err)
{
   int i, j, rj;
   struct ImageInfo *ptr;

   if ((!mf_first)||(!mf_data)||(!mf_info)) return MPK_NO_DATA;
   if (mf_refstar<=0) return MFI_REF_NOT_FOUND;
   if ((star<=0)||(!jd)||(!mag)||(!err)) return MPK_INVALID_PAR;

   ptr = mf_first;
   i = 0;
   j = star-1;
   rj = mf_refstar-1;
   while (ptr) {
	   jd[i]  = ptr->jd;
	   if (MAG(i,j)<99 && MAG(i,rj)<99) {
	     mag[i] = MAG(i,j) - MAG(i,rj);
	     err[i] = sqrt(ERR(i,j)*ERR(i,j)+ERR(i,rj)*ERR(i,rj));
       } else {
	     mag[i] = 99.999;
	     err[i] = 9.9999;
       }
	   ptr = ptr->next;
	   i++;
   }

   return 0;
}

int mfind_write(const char *outfile)
{
  FILE  *outf;
  char   msg[MAXLINE];
  int    j;

  PrintOut(0, "---------------------------------------------");
  sprintf(msg,"Output file:                        %s",outfile);
  PrintOut(0, msg);

  /* Write output to a file */
  outf = fopen(outfile,"w+");
  if (outf==NULL) return MPK_CANNOT_OPEN_OUT;

  fprintf(outf,"INDEX MEAN_MAG STDEV GOODPOINTS\n");
  fprintf(outf,"Reference star: %d\n",mf_refstar);

  for (j=0;j<mf_nstars;j++) {
	  if (mf_info[j].valid) {
	  	fprintf(outf," %6d %8.4f %8.4f %6d\n", j+1, mf_info[j].mean, mf_info[j].stdev, mf_info[j].npoints);
  	  } else {
	  	fprintf(outf," %6d %8.4f %8.4f %6d\n", j+1, 999.99, 99.999, 0);
      } 	
  }

  fclose(outf);

  /* Normal return */
  return 0;
}

int mfind_wdata(int star, const char *outfile)
{
  FILE  *outf;
  char   msg[MAXLINE];
  int    i, j, rj;
  double mag, err;
  struct ImageInfo *ptr;

  PrintOut(0, "---------------------------------------------");
  sprintf(msg,"Data file:                          %s",outfile);
  PrintOut(0, msg);

  if ((!mf_first)||(!mf_data)||(!mf_info)) return MPK_NO_DATA;
  if (mf_refstar<=0) return MFI_REF_NOT_FOUND;
  if ((star<=0)||(!outfile)) return MPK_INVALID_PAR;

  /* Write output to a file */
  outf = fopen(outfile,"w+");
  if (outf==NULL) return MPK_CANNOT_OPEN_OUT;

  fprintf(outf,"JD MAG ERR\n");
  fprintf(outf,"Reference star: %d\n",mf_refstar);

  ptr = mf_first;
  i = 0;
  j = star-1;
  rj = mf_refstar-1;
  while (ptr) {
	   if ((MAG(i,j)<99)&&(MAG(i,rj)<99)) {
	     mag = MAG(i,j) - MAG(i,rj);
	     err = sqrt(ERR(i,j)*ERR(i,j)+ERR(i,rj)*ERR(i,rj));
       } else {
	     mag = 99.999;
	     err = 9.9999;
       }
 	   fprintf(outf,"%13.5f %8.4f %8.4f\n", ptr->jd, mag, err);
	   ptr = ptr->next;
	   i++;
   }

  /* Normal return */
  fclose(outf);
  return 0;
}

int mfind_clean()
/* Closes output file and frees allocated memory */
{
  struct ImageInfo *ptr, *next;
  struct MagInfo *dta, *dnext;

  if (mf_data) free(mf_data);
  if (mf_info) free(mf_info);
  mf_data = NULL;
  mf_info = NULL;

  if (mf_first) {
 	ptr = mf_first;
	while (ptr) {
		next = ptr->next;
		dta  = ptr->first;
		while (dta) {
			dnext = dta->next;
			free(dta);
			dta = dnext;
		}
		free(ptr);
		ptr = next;
	}
  }
  mf_first = mf_last = NULL;
  mf_nimages = mf_nstars = 0;

  // Normal return
  return 0;
}
