/**************************************************************

mparams.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mparams.h"
#include "cmunipack.h"
#include "../common/comfun.h"

//-------------------------   CONSTANTS AND DATA TYPES   ----------------------------

// Tokens:
#define	LEX_ERROR	-1
#define LEX_NONE	0
#define LEX_END		1
#define LEX_VAL		2
#define	LEX_ID		3
#define LEX_DIV		4
#define LEX_EQ		5
#define LEX_SP		6

// Lexical analyser state:
struct LexData {
	char	*in;
	char 	*inptr;
	char 	*buf;
	char 	*bufptr;
	int  	given_back;
	int     eof;
};

struct ParserData {
	char	*id;
	parser_callback *cbproc;
};

// Forward declarations:
static int A(struct LexData *x, struct ParserData *y);
static int B(struct LexData *x, struct ParserData *y);
static int C(struct LexData *x, struct ParserData *y);
static int D(struct LexData *x, struct ParserData *y);

//-------------------------   LEXICAL ANALYSER   ----------------------------

static void lex_init(const char *in, struct LexData *lex)
// Set initial state of lexical analysis
{
	size_t len = strlen(in);
	lex->in = (char*) malloc(len+1);
	strcpy(lex->in, in);
	lex->inptr = lex->in;
	lex->buf = (char*) malloc(len+1);
	lex->bufptr = lex->buf;
	lex->given_back = LEX_NONE;
	lex->eof = 0;
}

static void lex_close(struct LexData *lex)
// Set initial state of lexical analysis
{
	free(lex->in);
	lex->in = lex->inptr = NULL;
	free(lex->buf);
	lex->buf = lex->bufptr = NULL;
}

static void lex_back(struct LexData *lex, int token)
// Give a token back
{
	lex->given_back = token;
}

static int lex(struct LexData *lex)
// Lexical analysis of star selection definition string
{
	int token, state;
	char ch;

	if (lex->given_back!=LEX_NONE) {
		token = lex->given_back;
		lex->given_back = LEX_NONE;
		return token;
	}
	lex->bufptr = lex->buf;
	lex->bufptr[0] = '\0';
	
	if (lex->eof) 
		return LEX_END;

	state = 0;		
	while ((ch=*lex->inptr++)!='\0') {
		switch (state)
		{
		case 0:
			// Initial state
			if (ch>='A' && ch<='Z') {
				state = 1; 
				*lex->bufptr++ = ch + ('a'-'A');
			} else
			if (ch>='a' && ch<='z') {
				state = 1; 
				*lex->bufptr++ = ch;
			} else
			if (ch>='0' && ch<='9') {
				state = 3; 
				*lex->bufptr++ = ch;
			} else
			if (ch=='+' || ch=='-') {
				state = 2; 
				*lex->bufptr++ = ch;
			} else
			if (ch=='"') {
				state = 5;
			} else
			if (ch==',') {
				return LEX_DIV;
			} else
			if (ch=='=') {
				return LEX_EQ;
			} else
			if (ch==' ' || ch=='\n' || ch=='\t' || ch=='\r') {
				state = 4; 
			} else
				return LEX_ERROR;
			break;
			
		case 1:
			// Identifier
			if (ch>='A' && ch<='Z') {
				*lex->bufptr++ = ch + ('a'-'A');
			} else
			if ((ch>='a' && ch<='z') || (ch>='0' && ch<='9') || (ch=='_')) {
				*lex->bufptr++ = ch;
			} else {
				lex->inptr--;
				*lex->bufptr = '\0';
				return LEX_ID;
			}
			break;
			
		case 2:
			// Sign before number
			if (ch>='0' && ch<='9') {
				*lex->bufptr++ = ch;
				state = 3;
			} else {
				return LEX_ERROR;
			}
			break;
			
		case 3:
			// Integer number, FP number or coordinate
			if ((ch>='0' && ch<='9') || (ch=='.') || (ch=='e') || (ch=='E') || (ch==':')) {
				*lex->bufptr++ = ch;
			} else {
				lex->inptr--;
				*lex->bufptr = '\0';
				return LEX_VAL;
			}
			break;
			
		case 4:
			// White spaces
			if (ch==' ' || ch=='\n' || ch=='\t' || ch=='\r') {
				state = 4; 
			} else {
				lex->inptr--;
				return LEX_SP;
			}
			break;
		
		case 5:
			// String constant
			if (ch=='"') {
				*lex->bufptr = '\0';
				return LEX_VAL;
			} else
			if (ch=='\\') {
				state = 6;
			} else {
				*lex->bufptr++ = ch;
			}
			break;
			
		case 6:
			// Escape character
			if (ch=='\\' || ch=='"') {
				*lex->bufptr++ = ch;
				state = 5;
			} else {
				return LEX_ERROR;
			}
			break;
			
		default:
			// Undefined state
			return LEX_ERROR;
		}
	}
	
	lex->eof = 1;
	if (state==1) {
		*lex->bufptr = '\0';
		return LEX_ID;
	}
	if (state==3) {
		*lex->bufptr = '\0';
		return LEX_VAL;
	}
	return LEX_END;
}

//-------------------------   SYNTAX ANALYSER   ----------------------------

// Starting procedure:
static int S(struct LexData *x, struct ParserData *y)
{
	switch (lex(x))
	{
	case LEX_ID:
		// S -> id A
		strcpy(y->id, x->buf);
		return A(x,y);
	}
	return MLS_SYNTAX_ERROR;
}

static int A(struct LexData *x, struct ParserData *y)
{
	int tok;
	
	switch (lex(x))
	{
	case LEX_SP:
		// A -> sp eq B
		tok = lex(x);
		if (tok!=LEX_EQ) 
			return MLS_SYNTAX_ERROR;
	case LEX_EQ:
		// A -> eq B
		return B(x,y);
	}
	return MLS_SYNTAX_ERROR;
}

static int B(struct LexData *x, struct ParserData *y)
{
	int tok;
	
	switch (lex(x))
	{
	case LEX_SP:
		// B -> sp int C
		tok = lex(x);
		if (tok!=LEX_VAL) 
			return MLS_SYNTAX_ERROR;
	case LEX_VAL:
		// B -> int C
		if (y->cbproc)
			y->cbproc(y->id, x->buf);
		return C(x,y);
	}
	return MLS_SYNTAX_ERROR;
}			

static int C(struct LexData *x, struct ParserData *y)
{
	int tok;
	
	tok = lex(x);
	switch (tok)
	{
	case LEX_SP:
		// C -> sp D
		return D(x,y);
	default:
		// C -> D
		lex_back(x, tok);
		return D(x,y);
	}
}

static int D(struct LexData *x, struct ParserData *y)
{
	switch (lex(x))
	{
	case LEX_DIV:
		// D -> , B
		return B(x,y);
	case LEX_ID:
		// D -> id A
		strcpy(y->id, x->buf);
		return A(x,y);
	case LEX_END:
		// D -> end
		return 0;
	}
	return MLS_SYNTAX_ERROR;
}

static void init(const char *params, struct ParserData *parser)
{
	parser->id = (char*) malloc(strlen(params)+1);
	if (parser->id)
		parser->id[0] = '\0';
	parser->cbproc = NULL;
}

static void close(struct ParserData *parser)
{
	if (parser->id)
		free(parser->id);
	parser->id = NULL;
}

//-------------------------   PARSER FUNCTION   ----------------------------

int cmpack_parse_params(const char *params, parser_callback *cbproc)
// Syntactic analysis for definition string
{
	int retval;
	struct LexData ld;
	struct ParserData pd;
	char txt[128];

	init(params, &pd);
	pd.cbproc = cbproc;
	lex_init(params, &ld);
	retval = S(&ld, &pd);
	if (retval==MLS_SYNTAX_ERROR) {
		int col = (int)(ld.inptr - ld.in)+1;
		sprintf(txt, "Syntax error on line 1 column %d", col);
		PrintOut(0, txt);
	}
	lex_close(&ld);
	close(&pd);
	return retval;
}
