/**************************************************************

find.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "comfun.h"
#include "phot.h"
#include "phfun.h"
#include "find.h"

#define D(x,y)    d[((x)-1)+(((y)-1)*ncol)]
#define G(x,y)    g[((x)-1)+(((y)-1)*nbox)]
#define H(x,y)    h[((x)-1)+(((y)-1)*ncol)]
#define SKIP(x,y) skip[((x)-1)+(((y)-1)*nbox)]

int Find(void)
// Find stars
{
  int   ncol, nrow, lastcl, lastro, nhalf, nbox, middle, pixels, i, j, ix, jx, iy, jy, n, kx, ky, failed;
  double *d;     /* image */
  double *g;     /* subarray with gaussian curve */
  double *h;     /* central heights of the best fitting Gaussian function for each pixel */
  char *skip;   /* which pixels will be skipped */
  double radius, radius2, wt, p, sigsq, sumg, sumgsq, jsq, rsq, denom, sgop, relerr, sgd, sd, sg, sgsq;
  double skymod, readns, phpadu, hmin, datum, dat, height, sharpness, temp;
  double sumgd, sumd, sdgdx, sdgdxs, sddgdx, sgdgdx, dgdx, hx, skylvl, dx, xcen, hy, dy, ycen, roundness;
  StarType *rec;

  /* Kontroly parametru */
  ncol  = ph_image.ncol;
  nrow  = ph_image.nrow;
  d     = ph_image.ccd;
  skymod= ph_image.skymod;
  if (ph_fravg<=0) {
	  PrintOut(0, "Number of frames averaged must be greater than zero.");
	  return MPK_INVALID_PAR;
  }
  if (ph_frsum<=0) {
	  PrintOut(0, "Number of frames summed must be greater than zero.");
	  return MPK_INVALID_PAR;
  }

  /* SECTION 1 - PREPARE SUBARRAY WITH CIRCULAR GAUSSIAN FUNCTION */

  radius = fmax(2.001,0.637*ph_fwhm);
  nhalf  = (int)radius;             /* number of pixels between the central pixel (exclusive)
                                       and the edge of the box (inclusive) */
  nbox   = 2*nhalf + 1;             /* length of the side of the subarray */
  middle = nhalf + 1;               /* middle point in the subarray */
  radius2 = radius*radius;			/* square radius */

  g = (double*) malloc(nbox*nbox*sizeof(double));
  h = (double*) malloc(nrow*ncol*sizeof(double));
  skip = (char *) malloc(nbox*nbox*sizeof(char));

  lastro = nrow - nhalf;           /* last row */
  lastcl = ncol - nhalf;           /* last column */

  sigsq  = (ph_fwhm/2.35482)*(ph_fwhm/2.35482);
  sumg = 0.0;                  /* sum of the values of the Gaussian function */
  sumgsq = 0.0;                /* sum of the squares of the values of the Gaussian function */
  pixels = 0;                  /* number of pixels in the circular area */
  for (j=1;j<=nbox;j++) {
     jsq = (j - middle)*(j - middle);
     for (i=1;i<=nbox;i++) {
        rsq = (i - middle)*(i - middle) + jsq;
        G(i,j) = exp(-0.5*rsq/sigsq);
        if (rsq <= radius2) {
           SKIP(i,j) = 0;
           sumg += G(i,j);
           sumgsq += G(i,j)*G(i,j);
           pixels++;
        } else {
           SKIP(i,j) = 1;
        }
     }
  }
  denom = sumgsq - (sumg*sumg)/pixels;    /* denominator of the fraction defining h */
  sgop = sumg/pixels;                     /* contains [G]/n */

  /* Urceni ruznych parametru */
  relerr = sqrt(1.0/denom);
  readns = ph_readns*sqrt(1.0*ph_frsum/ph_fravg);
  phpadu = ph_gain*ph_fravg;
  hmin   = ph_thresh*relerr*sqrt(readns*readns + fmax(0.0,skymod)/phpadu);

  ph_image.relerr = relerr;
  ph_image.phpadu = phpadu;
  ph_image.readns = readns;
  ph_image.relerr = relerr;
  ph_image.thres  = hmin;

  ClearStarList();

  /* SECTION 2 - Compute central heights of the best fitting Gaussian function */

  for (jy=1;jy<=nrow;jy++) {
	  for (jx=1;jx<=ncol;jx++) {
		  sgd = 0.0;
		  sd  = 0.0;
		  sgsq = sumgsq;
		  sg  = sumg;
		  p   = pixels;

		  /* Compute the local brightness enhancement for the pixel [jx,jy] */
		  for (ix=jx-nhalf;ix<=jx+nhalf;ix++) {
			  i = middle+(ix-jx);
			  for (iy=jy-nhalf;iy<=jy+nhalf;iy++) {
				  j = middle+(iy-jy);
				  if (!SKIP(i,j)) {
					  if ((1<=ix)&&(ix<=ncol)&&(1<=iy)&&(iy<=nrow)&&(D(ix,iy)>ph_minvalue)&&(D(ix,iy)<ph_maxvalue)) {
						  dat = D(ix,iy);
                          sgd  += G(i,j)*dat;
                          sd   += dat;
                      } else {
                          sgsq -= G(i,j)*G(i,j);
                          sg   -= G(i,j);
                          p--;
                      }
                  }
              }
          }

          if (p>1) {
	           /* Compute central height of the best fitting Gaussian function */
	           if (p<pixels) {
		           sgsq = sgsq - (sg*sg)/p;
                   if (sgsq!=0.0) sgd = (sgd - sg*sd/p)/sgsq;
                             else sgd = 0.0;
               } else {
                   sgd = (sgd - sgop*sd)/denom;
               }
           } else {
               sgd = 0.0;
           }
           H(jx,jy) = sgd;   /* central height of the best fitting Gaussian function */
       }
  }

  SKIP(middle,middle) = 1;
  pixels--;

  /* SECTION 3 - Search for local maxima in the convolved brightness data */

  for (jy=middle;jy<=lastro;jy++) {
	  for (jx=middle;jx<=lastcl;jx++) {

		  /* Test the height: must be above a threshold */
		  height = H(jx,jy);
          if (height<hmin) continue;  /* Test failed */

  		  /* the vicinity of this pixel: must be above each pixel within a radius equal to 1.5 sigma */
  		  failed = 0;
          for (ix=jx-nhalf;ix<=jx+nhalf;ix++) {
	          if ((1<=ix)&&(ix<=ncol)) {
		          for (iy=jy-nhalf;iy<=jy+nhalf;iy++) {
			          if ((1<=iy)&&(iy<=nrow)) {
		                  i = middle+(ix-jx);
  			              j = middle+(iy-jy);
			              if (!SKIP(i,j)) {
				              if (height<H(ix,iy)) { failed = 1; break; }
			              }
			          }
		          }
	          }
	          if (failed) break;
          }
          if (failed) continue;  /* Test failed - try next pixel */

  	      /* Derive the shape indices */
          sharpness = 0.0;
          datum = D(jx,jy);
          if ((datum>ph_minvalue)&&(datum<ph_maxvalue)) {
	          p = 0.0;
	          for (ix=jx-nhalf;ix<=jx+nhalf;ix++) {
		          if ((1<=ix)&&(ix<=ncol)) {
			          i=middle+(ix-jx);
			          temp = 0.0;
			          for (iy=jy-nhalf;iy<=jy+nhalf;iy++) {
				          j = middle+(iy-jy);
				          if ((!SKIP(i,j))&&(1<=iy)&&(iy<=nrow)) {
					          dat = D(ix,iy);
					          if ((dat>ph_minvalue)&&(dat<ph_maxvalue)) {
						          temp += (dat - skymod);
						          p    += 1.0;
					          }
				          }
			          }
                      sharpness += temp;
                  }
              }
              sharpness = (D(jx,jy)-skymod-sharpness/p)/height;
              if ((sharpness<ph_shrplo)||(sharpness>ph_shrphi)) continue;    /* Test failed - try next pixel */
          }

          /* SKIP ALL THIS IF THE STAR IS TOO NEAR THE EDGE OF THE FRAME
          if ((jx<middle)||(jx>lastcl)||(jy<middle)||(jy>lastro)) continue;*/

          ix = jx - middle;
          iy = jy - middle;

          sumgd = 0.0;                  /* sum of the Gaussian times the data */
          sumgsq = 0.0;                 /* sum of the Gaussian squared */
          sumg = 0.0;                   /* sum of the Gaussian */
          sumd = 0.0;                   /* sum of the data */
          sdgdx = 0.0;                  /* sum of d(Gaussian)/dx */
          sdgdxs = 0.0;                 /* sum of {d(Gaussian)/dx}^2 */
          sddgdx = 0.0;                 /* sum of data times d(Gaussian)/dx */
          sgdgdx = 0.0;                 /* sum of Gaussian times d(Gaussian)/dx */
          p = 0.0;
          n = 0;

          for (i=1;i<=nbox;i++) {
	          sg = 0.0;                  /* temporary sum of the Gaussian */
              sd = 0.0;                  /* temporary sum of the data */
              kx = ix + i;
              for (j=1;j<=nbox;j++) {
	              wt = middle - fabs(j - middle);
                  ky = iy + j;
                  dat = D(kx,ky);
                  if ((dat > ph_minvalue)&&(dat < ph_maxvalue)) {
	                  sd += (dat - skymod)*wt;
	                  sg += G(i,j)*wt;
                  }
              }
              if (sg > 0.0) {
	              wt = middle - fabs(i - middle);
                  sumgd  += wt*sg*sd;
                  sumgsq += wt*sg*sg;
                  sumg   += wt*sg;
                  sumd   += wt*sd;
                  dgdx    = sg*(middle - i);
                  sdgdxs += wt*dgdx*dgdx;
                  sdgdx  += wt*dgdx;
                  sddgdx += wt*sd*dgdx;
                  sgdgdx += wt*sg*dgdx;
                  p      += wt;
                  n++;
              }
          }

          if (n<=2) continue;   /* we need at least three points */
          hx = (sumgd - sumg*sumd/p)/(sumgsq - (sumg*sumg)/p);
          if (hx<=0.0) continue;
          skylvl = (sumd - hx*sumg)/p;
          dx = (sgdgdx - (sddgdx - sdgdx*(hx*sumg + skylvl*p)))/(hx*sdgdxs/sigsq);
          /* compute the first-order correction to the x-centroid of the star */
          xcen = jx + dx/(1.0 + fabs(dx));
          if ((xcen<0.5)||(xcen>=ncol+0.5)) continue; /* if the best estimate of the star's center falls
                                                        outside the ph_image, reject it */

          /* Compute the height of the y-marginal Gaussian distribution */
          sumgd  = 0.0;
          sumgsq = 0.0;
          sumg   = 0.0;
          sumd   = 0.0;
          sdgdx  = 0.0;
          sdgdxs = 0.0;
          sddgdx = 0.0;
          sgdgdx = 0.0;
          p      = 0.0;
          n      = 0;
          for (j=1;j<=nbox;j++) {
	          ky = iy + j;
              sg = 0.0;
              sd = 0.0;
              for (i=1;i<=nbox;i++) {
	              wt = middle - fabs(i - middle);
	              kx = ix + i;
                  dat = D(kx,ky);
                  if ((dat > ph_minvalue)&&(dat < ph_maxvalue)) {
	                  sd += (dat - skymod)*wt;
                      sg += G(i,j)*wt;
                  }
              }
              if (sg>0.0) {
                  wt = middle - fabs(j - middle);
                  sumgd += wt*sg*sd;
                  sumgsq+= wt*sg*sg;
                  sumg  += wt*sg;
                  sumd  += wt*sd;
                  dgdx   = sg*(middle - j);
                  sdgdx += wt*dgdx;
                  sdgdxs+= wt*dgdx*dgdx;
                  sddgdx+= wt*sd*dgdx;
                  sgdgdx+= wt*sg*dgdx;
                  p     += wt;
                  n++;
              }
          }

          if (n<=2) continue;   /* we need at least three points */
          hy = (sumgd - sumg*sumd/p)/(sumgsq - (sumg*sumg)/p);
          if (hy<=0.0) continue;
          skylvl = (sumd - hy*sumg)/p;
          dy = (sgdgdx - (sddgdx - sdgdx*(hy*sumg + skylvl*p)))/(hy*sdgdxs/sigsq);
          ycen = jy + dy/(1.0 + fabs(dy));
          if ((ycen<0.5)||(ycen>nrow+0.5)) continue;    /* if the best estimate of the star's center falls
                                                        outside the ph_image, reject it */
          /* Derive roundness of object */
          roundness = 2.0*(hx - hy)/(hx + hy);
          if ((roundness<ph_rndlo)||(roundness>ph_rndhi)) continue;   /* Fails roundness test */

          /* Prints it's parameters to the monitor and output file */
          AddStar(&rec);
          rec->xmax      = jx;
          rec->ymax      = jy;
          rec->xcen      = xcen;
          rec->ycen      = ycen;
          rec->height    = datum;
          rec->sharpness = sharpness;
          rec->roundness = roundness;

          /* continue with the next pixel */
      }
  }

  /* Section 4 - frees alocated memory */
  free(g);
  free(h);
  free(skip);
  /* Normal return */

  return 0;
}
